/*---------------------------------------------------------------------------

	File:
		carnival.cpp
	
	Abstract:
		Implements the "Carnival" game driver functions.

	Revisions:
		09-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information derived from the excellent emulator, MAME by Nicola
	Salmoria.
	
---------------------------------------------------------------------------*/

#include "pch.h"
#include <stdio.h>
#include "misc.h"
#include "z80api.h"
#include "game.h"
#include "game1.h"

static BYTE _rgbChars[256][8][8];
static BYTE _rgbChars2[256][8][8];
static BYTE _rgbSprites[256][32][32];
static BYTE _rgbSpriteMask[256][32][32];

#include "blit.h"

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static unsigned char palette[] =
{
	0x00,0x00,0x00,	/* BLACK */
	0xdb,0x00,0x00,	/* RED */
	0xdb,0x92,0x49,	/* BROWN */
	0xff,0xb6,0xdb,	/* PINK */
	0x00,0xdb,0x00,	/* UNUSED */
	0x00,0xdb,0xdb,	/* CYAN */
	0x49,0xb6,0xdb,	/* DKCYAN */
	0xff,0xb6,0x49,	/* DKORANGE */
	0x88,0x88,0x88,	/* UNUSED */
	0xdb,0xdb,0x00,	/* YELLOW */
	0xff,0x00,0xdb,	/* UNUSED */
	0x24,0x24,0xdb,	/* BLUE */
	0x00,0xdb,0x00,	/* GREEN */
	0x49,0xb6,0x92,	/* DKGREEN */
	0xff,0xb6,0x92,	/* LTORANGE */
	0xdb,0xdb,0xdb	/* GREY */
};

enum {BLACK,RED,BROWN,PINK,UNUSED1,CYAN,DKCYAN,DKORANGE,
		UNUSED2,YELLOW,UNUSED3,BLUE,GREEN,DKGREEN,LTORANGE,GREY};

static const unsigned char colortable[] =
{
	0,1,2,3,4,5,6,7,
	0,2,3,4,5,6,7,8,
	0,3,4,5,6,7,8,9,
	0,4,5,6,7,8,9,10,
	0,5,6,7,8,9,10,11,
	0,6,7,8,9,10,11,12,
	0,7,8,9,10,11,12,13,
	0,8,9,10,11,12,13,14,
	0,9,10,11,12,13,14,15,
	0,10,11,12,13,14,15,1,
	0,11,12,13,14,15,1,2,
	0,12,13,14,15,1,2,3,
	0,13,14,15,1,2,3,4,
	0,14,15,1,2,3,4,5,
	0,15,1,2,3,4,5,6,
	0,15,1,2,3,4,5,6
};

static void BuildCompositePalette(PALETTEENTRY *ppeOut, UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < DIM(colortable); cpeOut++)
	{
		ppeOut[cpeOut].peRed   = palette[3 * colortable[cpeOut] + 0];
		ppeOut[cpeOut].peGreen = palette[3 * colortable[cpeOut] + 1];
		ppeOut[cpeOut].peBlue  = palette[3 * colortable[cpeOut] + 2];
	}
}

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	static const TCHAR szPath[] = "carnival";

	static const SIZE _sizBuffer = { 256 + 32, 256 + 32};
	static const RECT _rcDisplay = { 0, 0, 256 + 32, 256 + 32};

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(gli.ppe, gli.cpe);

	//
	//
	//
	
	static const RomModule2 ROMs[] =
	{
		{ "651u33.cpu", 0x0000, 0x0400, 0x0000 },
		{ "651u33.cpu", 0x0400, 0x0400, 0x4000 },
		{ "652u32.cpu", 0x0000, 0x0400, 0x0400 },
		{ "652u32.cpu", 0x0400, 0x0400, 0x4400 },
		{ "653u31.cpu", 0x0000, 0x0400, 0x0800 },
		{ "653u31.cpu", 0x0400, 0x0400, 0x4800 },
		{ "654u30.cpu", 0x0000, 0x0400, 0x0c00 },
		{ "654u30.cpu", 0x0400, 0x0400, 0x4c00 },
		{ "655u29.cpu", 0x0000, 0x0400, 0x1000 },
		{ "655u29.cpu", 0x0400, 0x0400, 0x5000 },
		{ "656u28.cpu", 0x0000, 0x0400, 0x1400 },
		{ "656u28.cpu", 0x0400, 0x0400, 0x5400 },
		{ "657u27.cpu", 0x0000, 0x0400, 0x1800 },
		{ "657u27.cpu", 0x0400, 0x0400, 0x5800 },
		{ "658u26.cpu", 0x0000, 0x0400, 0x1c00 },
		{ "658u26.cpu", 0x0400, 0x0400, 0x5c00 },
		{ "659u8.cpu", 0x0000, 0x0400,  0x200 },
		{ "659u8.cpu", 0x0400, 0x0400, 0x6000 },
		{ "660u7.cpu", 0x0000, 0x0400,  0x240 },
		{ "660u7.cpu", 0x0400, 0x0400, 0x6400 },
		{ "661u6.cpu", 0x0000, 0x0400,  0x280 },
		{ "661u6.cpu", 0x0400, 0x0400, 0x6800 },
		{ "662u5.cpu", 0x0000, 0x0400,  0x2c0 },
		{ "662u5.cpu", 0x0400, 0x0400, 0x6c00 },
		{ "663u4.cpu", 0x0000, 0x0400,  0x300 },
		{ "663u4.cpu", 0x0400, 0x0400, 0x7000 },
		{ "664u3.cpu", 0x0000, 0x0400,  0x340 },
		{ "664u3.cpu", 0x0400, 0x0400, 0x7400 },
		{ "665u2.cpu", 0x0000, 0x0400,  0x380 },
		{ "665u2.cpu", 0x0400, 0x0400, 0x7800 },
		{ "666u1.cpu", 0x0000, 0x0400,  0x3c0 },
		{ "666u1.cpu", 0x0400, 0x0400, 0x7c00 },

		{ NULL }
	};
	
	if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	static const MY_GFX_DECODE mgdChars1 =
	{
		{ 8, 8 },
		40,
		8 * 8,
		1,
		0,
		1,
	{ 0, 1, 2, 3, 4, 5, 6, 7 },	/* pretty straightforward layout */
	{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8, 8*8, 9*8 },
	};

	MyGfxDecode(RAM + 0x01E8, mgdChars1, (LPBYTE)_rgbChars);

	static const MY_GFX_DECODE mgdChars2 =
	{
		{ 8, 8 },
		256,
		8 * 8,
		1,
		0,
		1,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8, 8*8, 9*8 },
		{ 7, 6, 5, 4, 3, 2, 1, 0 },	/* pretty straightforward layout */
	};

	MyGfxDecode(RAM + 0x0400, mgdChars2, (LPBYTE)_rgbChars2);

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	return RAM[A];
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int vblank;

static int Interrupt(void)
{
	if (!CheckMessages())
		CPURunning = 0;

	Throttle(60); // FRAMES_PER_SECOND);

	/* let carnival_IN1_r() know that it is time to report a vblank */
	vblank = 1;

	return IGNORE_INT;
}

static BYTE In(BYTE bAddr)
{
	switch (bAddr)
	{
		case 0x01:
		{
			int res;

			res = 0x00;

			/* I'm not yet sure about how the vertical blanking should be handled. */
			/* I think that IN1_VBLANK should be 1 during the whole vblank, which */
			/* should last roughly 1/12th of the frame. */
			if (vblank && ICount > 50000 - (50000 / 12))
				res |= (1 << 3); // IN1_VBLANK;
			else vblank = 0;

			return res;
		}

		default:
			return 0;
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

// TODO:  call higher-level Exec function that handles the common parts
//        of Interrupt, pass it our CPU speed and frame rate

static HRESULT Run(LPVOID pvContext)
{
	EXEC_Z80_STRUCT es;
	
	ZERO_INIT(es);
	
	es.nCallbackFreq = 50000; // CALLBACK_FREQ;
	es.pfnInterrupt = Interrupt;
	es.pfnReadMemory = M_RDMEM;
	es.pfnWriteMemory = M_WRMEM;
	es.rgbMemory = RAM;
	es.pfnIn = In;
	
	return ExecZ80(es);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static HRESULT RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

	if (GetAsyncKeyState('G') < 0)
	{
		if (GetAsyncKeyState('1') < 0)
		{
			for (i = 0; i < 256; i++)
				DrawChar(rgbBuffer, lPitch, 8 * (i % 32), 8 * (i / 32), i, 0);
		}

		return S_OK;
	}

	const BOOL fShowI = GetAsyncKeyState('I') < 0;
	const BOOL fShow0 = GetAsyncKeyState('0') < 0;

	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 32; j++)
		{
			DrawChar((LPBYTE)_rgbChars, RAM[0xE000 + i + j * 32], 0,
				8 * (31 - j), 8 * i, rgbBuffer, lPitch);
		}
	}

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdCarnival = 
{
	TEXT("Carnival"),
	Init,
	Run,
	RenderScene,
	NULL,
	NULL,
	NULL,
	NULL
};

