/*---------------------------------------------------------------------------
	
	File:
		galaga.cpp
	
	Abstract:
		Implements the "Galaga" game driver functions.

	Revisions:
		10-May-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information for this game was generously provided by Martin
	Scragg.  DIP switch code adopted from MAME by Nicola Salmoria (without
	which the game wouldn't get past the boot sequence).  Credits to Joe
	Massey for starry background algorithm.

	TODO:  Patch ROMs w/NOPs to implement busy loop performance enhancements
	while keeping the CPU code generic.

	We may be able to keep the ROM check from failing by making the patched
	area's use a callback function for reading from them instead of direct
	access.  In the callback, just return the expected values.

	when the main rom boots, it writes 0x10 to 0x7100
	sets interrupt mode 1
	zero's memory from 0x99E0 to 0x99F0
	writes zero to 6823 (hold cpus 2 and 3 in reset)
	writes one to 6822 (disable NMI on cpu 3)
	disables interrupts
	writes one to 6830


	7100 - custom I/O chip interface
	99E0 -
	6823 -
	6822 -
	6830

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

void GalagaUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer, const BYTE *RAM,
	BOOL &_fSoundChanged);

static const int _rgnGalagaStarScrollSpeed[] = 
{
	-2, 6, -6, 2, -4, 4, 0, 0 
};

static BOOL _fSwitchReadMode;
static int _cCredits;

static BOOL _fSoundChanged = TRUE; // NOTE:  force calculation first time

#define SOUND_BASE 0x6800

struct GAME_RESOURCES
{
	DECODE_GRAPHICS rgGraphicDecodeInfo[16]; // gathered from script
	IMAGE rgImages[16]; // 
};

struct GALAGA
{
	IExecZ80 *pZ80[3];
	IMAGE *pimgChars;
	IMAGE *pimgSprites;
	BYTE *pVideoRam;
	BYTE *pColorRam;
	BYTE *rgpIrqEnable[3];
	BOOL fNmi;
};

static GALAGA _galaga =
{
	{ NULL, NULL, NULL },
	NULL,
	NULL,
	&RAM[0x8000],
	&RAM[0x8400],
	{ &RAM[0x6820], &RAM[0x6821], &RAM[0x6822] }
};

static const TCHAR _szPath[] = "galagabl";

RAGE_ROMTABLE_BEGIN(_rgMainRoms)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_1", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_2", 0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_3", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_4", 0x3000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgSecondRoms)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_5", 0x0000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgAudioRoms)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_7", 0x0000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgCharacterRoms)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_8", 0x0000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgSpriteRoms)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_A", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("GALAGABL.1_9", 0x1000, 0x1000)
RAGE_ROMTABLE_END

static const SIZE _sizBuffer = { 256 + 16, 256 + 32 + 16};
static const RECT _rcDisplay = { 16, 0, 256 - 16, 256 + 32 };

static BYTE _rgbSecondROM[0x4000];
static BYTE _rgbAudioROM[0x4000];

static const PALETTEENTRY GalagaPalette[] =
{
	{ 0x00, 0x00, 0x00 },		// 0
	{ 0x00, 0x00, 0x00 },		// 1
	{ 0x00, 0x00, 0xFF },		// 2
	{ 0xB6, 0x00, 0xFF },		// 3
	{ 0x00, 0x6D, 0xFF },		// 4
	{ 0x00, 0x00, 0x00 },		// 5
	{ 0x00, 0xFF, 0x00 },		// 6
	{ 0x00, 0x00, 0x00 },		// 7
	{ 0x00, 0xB6, 0xFF },		// 8
	{ 0x00, 0xFF, 0xFF },		// 9
	{ 0xFF, 0x00, 0xFF },		// 10
	{ 0x00, 0x00, 0x00 },		// 11
	{ 0xFF, 0x92, 0x00 },		// 12
	{ 0xFF, 0xFF, 0x00 },		// 13
	{ 0xFF, 0x00, 0x00 },		// 14
	{ 0xDB, 0xDB, 0xFF },		// 15
	{ 0x00, 0x00, 0x00 },		// 16
	{ 0x00, 0x92, 0xAA },		// 17
	{ 0x00, 0x00, 0xFF },		// 18
	{ 0x92, 0x00, 0xFF },		// 19
	{ 0x00, 0x6D, 0xFF },		// 20
	{ 0x24, 0x92, 0x00 },		// 21
	{ 0x00, 0xFF, 0x00 },		// 22
	{ 0xDB, 0x49, 0x00 },		// 23
	{ 0xB6, 0xB6, 0xFF },		// 24
	{ 0x00, 0xFF, 0xFF },		// 25
	{ 0xFF, 0x00, 0xFF },		// 26
	{ 0xFF, 0xB6, 0x00 },		// 27
	{ 0xFF, 0x92, 0x00 },		// 28
	{ 0xFF, 0xFF, 0x00 },		// 29
	{ 0xFF, 0x00, 0x00 },		// 30
	{ 0xDB, 0xDB, 0xFF }		// 31
};

static const BYTE GalagaColorTable[] =
{
	// characters...
	0x0F, 0x00, 0x00, 0x06,
	0x0F, 0x0D, 0x01, 0x00,
	0x0F, 0x02, 0x0C, 0x0D,
	0x0F, 0x0B, 0x01, 0x00,
	0x0F, 0x01, 0x00, 0x01,
	0x0F, 0x00, 0x00, 0x02,
	0x0F, 0x00, 0x00, 0x03,
	0x0F, 0x00, 0x00, 0x05,
	0x0F, 0x00, 0x00, 0x09,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x0F, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x0F, 0x0B, 0x07, 0x06,
	0x0F, 0x06, 0x0B, 0x07,
	0x0F, 0x07, 0x06, 0x0B,
	0x0F, 0x0F, 0x0F, 0x01,
	0x0F, 0x0F, 0x0B, 0x0F,
	0x0F, 0x02, 0x0F, 0x0F,
	0x0F, 0x06, 0x06, 0x0B,
	0x0F, 0x06, 0x0B, 0x0B,

	// sprites...
	0x0F, 0x08, 0x0E, 0x02,
	0x0F, 0x05, 0x0B, 0x0C,
	0x0F, 0x00, 0x0B, 0x01,
	0x0F, 0x01, 0x0B, 0x02,
	0x0F, 0x08, 0x0D, 0x02,
	0x0F, 0x06, 0x01, 0x04,
	0x0F, 0x09, 0x01, 0x05,
	0x0F, 0x07, 0x0B, 0x01,
	0x0F, 0x01, 0x06, 0x0B,
	0x0F, 0x01, 0x0B, 0x00,
	0x0F, 0x01, 0x02, 0x00,
	0x0F, 0x00, 0x01, 0x06,
	0x0F, 0x00, 0x00, 0x06,
	0x0F, 0x03, 0x0B, 0x09,
	0x0F, 0x06, 0x02, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

#include "stars.h"

static STARS _stars;

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);
static BYTE M_RDMEMSecond(WORD A);
static void M_WRMEMSecond(WORD A, BYTE);
static BYTE M_RDMEMAudio(WORD A);
static void M_WRMEMAudio(WORD A, BYTE);

static void BuildCompositePalette(PALETTEENTRY *ppeOut, UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < DIM(GalagaColorTable); cpeOut++)
		ppeOut[cpeOut] = GalagaPalette[31 - GalagaColorTable[cpeOut]];
}

static const DECODE_GRAPHICS dgsChars =
{
	0x00000001,
	{
		{ 8, 8 },
		128,
		16 * 8,
		2,
		{ 0, 4 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 8*8+0, 8*8+1, 8*8+2, 8*8+3, 0, 1, 2, 3 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgsSprites =
{
	0x00000001,
	{
		{ 16, 16 },
		128,
		64 * 8,
		2,
		{ 0, 4 },
		2,
		{ 39 * 8, 38 * 8, 37 * 8, 36 * 8, 35 * 8, 34 * 8, 33 * 8, 32 * 8, 7 * 8, 6 * 8, 5 * 8, 4 * 8, 3 * 8, 2 * 8, 1 * 8, 0 * 8 },
		{ 0, 1, 2, 3, 8*8, 8*8+1, 8*8+2, 8*8+3, 16*8+0, 16*8+1, 16*8+2, 16*8+3, 24*8+0, 24*8+1, 24*8+2, 24*8+3 },
	},

	1,
	
	4, 
	
	{
		0 * 128,
		1 * 128,
		2 * 128,
		3 * 128
	},
	0, 
	NULL
};

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
#else
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
	{ 0x0000, 0x3FFF, 0, NULL, M_WRMEM },
	{ 0x8000, 0x9FFF, 0, NULL, NULL },
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

static const EXECZ80_MEM_DESCRIPTOR rgmdSecond[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEMSecond, M_WRMEMSecond },
#else
	{ 0x0000, 0xFFFF, 0, NULL, NULL },
	{ 0x4000, 0xFFFF, 0, M_RDMEMSecond, M_WRMEMSecond },
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

static const EXECZ80_MEM_DESCRIPTOR rgmdAudio[] = 
{
#if 0
	{ 0x0000, 0xFFFF, 0, M_RDMEMAudio, M_WRMEMAudio },
#else
	{ 0x0000, 0xFFFF, 0, NULL, NULL },
	{ 0x4000, 0xFFFF, 0, M_RDMEMAudio, M_WRMEMAudio },
#endif
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

HRESULT GalagaInit(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(gli.ppe, gli.cpe);

	if (readroms(RAM, _rgMainRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(_rgbSecondROM, _rgSecondRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	if (readroms(_rgbAudioROM, _rgAudioRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	BYTE tmpstorage[0x2000];

	//
	//  Decode characters
	//
	
	if (readroms(tmpstorage, _rgCharacterRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	MyGfxDecode3(dgsChars, tmpstorage, _galaga.pimgChars);

	//
	//  Decode sprites
	//
	
	if (readroms(tmpstorage, _rgSpriteRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	MyGfxDecode3(dgsSprites, tmpstorage, _galaga.pimgSprites);

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory   = RAM;
	cs.pfnMemRead  = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd         = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _galaga.pZ80[0]);

	if (FAILED(hr))
		return hr;

	//
	//
	//
	
	ZERO_INIT(cs);

	cs.rgbMemory   = _rgbSecondROM;
	cs.pfnMemRead  = M_RDMEMSecond;
	cs.pfnMemWrite = M_WRMEMSecond;
	cs.pmd         = rgmdSecond;
	
	hr = EclCreateZ80(cs, _galaga.pZ80[1]);

	if (FAILED(hr))
		return hr;

	//
	//
	//
	
	ZERO_INIT(cs);

	cs.rgbMemory   = _rgbAudioROM;
	cs.pfnMemRead  = M_RDMEMAudio;
	cs.pfnMemWrite = M_WRMEMAudio;
	cs.pmd         = rgmdAudio;
	
	hr = EclCreateZ80(cs, _galaga.pZ80[2]);

	if (FAILED(hr))
		return hr;

	//
	//
	//
	
	srand(timeGetTime());

	_stars.Init();
	
	// NOTE:  find matching color in lookup table...
	
	for (int i = 0; i < _stars.m_cStars; i++)
	{
		for (int j = 0; j < 32; j++)
		{
			if (_stars.m_rgStars[i].c == GalagaColorTable[j])
				_stars.m_rgStars[i].c = j;
		}
	}

	//
	//  Patch over ROM busy loops with HALT instructions.  These are handled
	//  specially in the Exec function.
	//

/*
	_rgbSecondROM[0x05B1] = 0x76; //  2nd cpu does a loop1: ld sp,9100h jp loop1
	_rgbAudioROM[0x00A7] = 0x76; // audio cpu does a loop1: jr loop1

	if (_rgbAudioROM[0x00A7] == 0x18) // opcode for jr
	if (_rgbAudioROM[0x00A8] == 0xFE) // -2 bytes

	if (_rgbSecondROM[0x05B1] == 31)
	if (_rgbSecondROM[0x05B2] == 00)
	if (_rgbSecondROM[0x05B3] == 91)
	if (_rgbSecondROM[0x05B4] == C3)
	if (_rgbSecondROM[0x05B5] == B1)
	if (_rgbSecondROM[0x05B6] == 05)
*/

	if (GetAsyncKeyState(VK_SHIFT) < 0)
	{
//		_rgbAudioROM[0x00A7] = 0x76; // audio cpu does a loop1: jr loop1
	}

	return S_OK;
}

void GalagaTerm()
{
	_galaga.pZ80[0]->Release();
	_galaga.pZ80[1]->Release();
	_galaga.pZ80[2]->Release();

	FreeImage(_galaga.pimgSprites);
	FreeImage(_galaga.pimgChars);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	if (A < 0x4000)
		return RAM[A];

	//
	//  Thanks again to MAME source... without the DIP switches, the game hangs
	//
	
	if (A >= 0x6800 && A <= 0x6807)
	{
		int bit0,bit1;

		int offset = A - 0x6800;

		bit0 = (0x97 >> offset) & 1;
		bit1 = (0xF7 >> offset) & 1;

		return bit0 | (bit1 << 1);
	}

	if (A == 0x7100) // read I/O chip status
		return 0x10;

	if (A >= 0x8000 && A <= 0x9FFF)
		return RAM[A];

	_ASSERT(FALSE);
	
	return RAM[A];
}

static BYTE M_RDMEMSecond(WORD A)
{
	if (A < 0x4000)
		return _rgbSecondROM[A];
		
	return M_RDMEM(A);
}

static BYTE M_RDMEMAudio(WORD A)
{
	if (A < 0x4000)
		return _rgbAudioROM[A];
		
	return M_RDMEM(A);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void WriteIoChip(WORD A, BYTE V)
{
//	TRACE("RAM[0x%04X]->0x%02X\n", A, V);

	switch (V)
	{
		case 0x10:
			break;

		case 0x61:
			_galaga.fNmi = TRUE; // generate an NMI to recover from DI/HALT
			break;

		case 0x71:
		{
			// hack!!!

			EXECZ80_CONTEXT ctx;

			_galaga.pZ80[0]->GetContext(ctx);

			BYTE *pDest = &RAM[ctx.DE2];

			static BOOL fCoin, fStart1, fStart2, fFire;
			static BOOL coin, start1, start2, fire;

			if (osd_key_pressed(OSD_KEY_3))
			{
				if (coin == 0 && _cCredits < 99)
					_cCredits++;
				
				coin = 1;
			}
			else
				coin = 0;

			/* check for 1 player start button */
			if (osd_key_pressed(OSD_KEY_1))
			{
				if (start1 == 0 && _cCredits >= 1)
					_cCredits--;
				
				start1 = 1;
			}
			else
				start1 = 0;

			/* check for 2 players start button */
			if (osd_key_pressed(OSD_KEY_2))
			{
				if (start2 == 0 && _cCredits >= 2)
					_cCredits -= 2;
				
				start2 = 1;
			}
			else
				start2 = 0;

			if (_fSwitchReadMode)
			{
				*pDest++ = 0x80;
			}
			else
			{
				*pDest++ = (_cCredits / 10) << 4 | (_cCredits % 10);
			}

			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			if ((b & 0x10) == 0)
			{
				if (fire)
					b |= 0x10;
				else
					fire = 1;
			}
			else
				fire = 0;

			*pDest++ = b;
			*pDest = b; // 0xFF;

			break;
		}

		case 0xA1: // goto switch read mode instead of credit mode
			_fSwitchReadMode = TRUE;
			break;

		case 0xA8:
			break;

		case 0xB1:
		{
			// hack!!!

			EXECZ80_CONTEXT ctx;

			_galaga.pZ80[0]->GetContext(ctx);

			RAM[ctx.DE2++] = 0;
			RAM[ctx.DE2++] = 0;
			RAM[ctx.DE2++] = 0;

			break;
		}

		case 0xE1: // goto credit mode instead of switch read mode
			_fSwitchReadMode = FALSE;
			break;

		default:
			_ASSERT(FALSE);
			break;
	}
}

static void M_WRMEM(WORD A, BYTE V)
{
	if (A >= 0x8000 && A <= 0x9FFF)
	{
		RAM[A] = V;
		return;
	}
	
	if (A >= SOUND_BASE && A < SOUND_BASE + 0x20)
	{
		RAM[A] = V & 0x0F;
		_fSoundChanged = TRUE;
		return;
	}

	if (A == 0x6830)
	{
		return;
	}

	if (A >= 0xA000 && A <= 0xA005)
	{
		RAM[A] = V;
	
		int scroll_index = RAM[0xA000] & 1;
	
		scroll_index = (scroll_index << 1) | RAM[0xA001] & 1;
		scroll_index = (scroll_index << 1) | RAM[0xA002] & 1;
		
		_stars.m_nScrollRate = _rgnGalagaStarScrollSpeed[scroll_index];

		return;
	}

	switch (A)
	{
		case 0x7100: // nasty little I/O chip
		{
			WriteIoChip(A, V);
			return;
		}

		case 0x6820:
		{
			RAM[A] = V;
			break;
		}

		case 0x6821:
		{
			RAM[A] = V;
			break;
		}

		case 0x6822: // inhibit cpu 3 NMI
		{
			RAM[A] = V;
			break;
		}

		case 0x6823: // allow cpus 2 and 3 to run
		{
			RAM[A] = V;
			break;
		}

		case 0x9A9B: // player ship explosion
		{
			RAM[A] = V;
			// SoundSetEnvelope(); 
//			V = 0;
			break;
		}

/*
		case 0xA007: // ???
		{
			RAM[A] = V;
			break;
		}

		case 0x2A3D: // ???
		{
			RAM[A] = V;
			break;
		}

		case 0x2A3F: // ???
		{
			RAM[A] = V;
			break;
		}
*/

		default:
//			_ASSERT(FALSE);
			break;
	}
}

static void M_WRMEMSecond(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

	M_WRMEM(A, V);
}

static void M_WRMEMAudio(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

	M_WRMEM(A, V);
}

/*---------------------------------------------------------------------------

	Exec

	Execute gameplay for one frame of Galaga.  I'm emulating all three CPUs
	in the system.  

	#1 - main: IRQ @ 60Hz, NMI
	#2 - secondary (controls half the sprites?): IRQ @ 60 Hz
	#3 - sound: NMI @ 120 Hz

	NOTE:  CPUs 2 and 3 are held in RESET while the memory at 0x6823 is zero.
	This condition only exists during the lengthy boot up process.  To speed
	up the booting, I've executed many times more t-states on the primary
	processor during each execution cycle.

---------------------------------------------------------------------------*/

int GalagaExec()
{
	int cCycles = 0;

	_stars.Exec();

	static const int cSlice = 3125000 / 60 / 2;
	
	if (RAM[0x6823]) // allow CPUs 2 and 3 to run?
	{
		EXECZ80_CONTEXT ctx;
		
		if (GetAsyncKeyState(VK_SHIFT) < 0)
		{
//			_rgbAudioROM[0x00A7] = 0x76; // audio cpu does a loop1: jr loop1
		}

/*
		_rgbSecondROM[0x05B4] = 0x76; //  nop
		_rgbSecondROM[0x05B5] = 0x18; //  jr 0x05B1 (-6 bytes)
		_rgbSecondROM[0x05B6] = 0xFA; //  
*/

		cCycles += _galaga.pZ80[1]->Exec(cSlice);

/*
		_galaga.pZ80[1]->GetContext(ctx);
		
		if (ctx.PC == 0x05B1 || ctx.PC == 0x05B4)
		{
			_rgbSecondROM[0x05B4] = 0x76; //  nop
			_rgbSecondROM[0x05B5] = 0x18; //  jr 0x05B1 (-6 bytes)
			_rgbSecondROM[0x05B6] = 0xFA; //  
		}
*/

		cCycles += _galaga.pZ80[2]->Exec(cSlice);

		_galaga.pZ80[2]->GetContext(ctx);
		
		if (ctx.PC == 0x00A7)
//		if (ctx.PC == 0x00A8)
		{
			ctx.PC == 0x00A7;

			_galaga.pZ80[2]->SetContext(ctx);
		}

		if (*_galaga.rgpIrqEnable[2] == 0)
			cCycles += _galaga.pZ80[2]->Nmi();

		cCycles += _galaga.pZ80[1]->Exec(cSlice);

		if (*_galaga.rgpIrqEnable[1] != 0)
			cCycles += _galaga.pZ80[1]->Irq(0xFF);

		cCycles += _galaga.pZ80[2]->Exec(cSlice);

		_galaga.pZ80[2]->GetContext(ctx);
		
		if (ctx.PC == 0x00A8)
		{
			ctx.PC == 0x00A7;

			_galaga.pZ80[2]->SetContext(ctx);
		}

		if (*_galaga.rgpIrqEnable[2] == 0)
			cCycles += _galaga.pZ80[2]->Nmi();
	}
	else
	{
		// special case...  allow CPU 1 to run very fast during bootup
		cCycles += _galaga.pZ80[0]->Exec(2 * cSlice * 60);
	}

	cCycles += _galaga.pZ80[0]->Exec(2 * cSlice);
	
	if (_galaga.fNmi)
	{
		_galaga.fNmi = FALSE;

		cCycles += _galaga.pZ80[0]->Nmi();
	}
	else
	{
		if (*_galaga.rgpIrqEnable[0] != 0)
			cCycles += _galaga.pZ80[0]->Irq(0xFF);
	}

	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void GalagaRenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;

	//
	//  Clear the buffer
	//
	//  CONSIDER:  have directdraw hardware fill
	//
	
	LPDWORD pDest = (LPDWORD)rgbBuffer + 4;

	for (i = 0; i < 256 + 32; i++)
	{
		memset(pDest, 0, 256 - 32);
		pDest += lPitch / 4;
	}

	//
	//  draw the faked starry background
	//
	
	_stars.Draw(rgbBuffer + lPitch * 16, lPitch);

	//
	//  Draw the 64 sprites
	//

	LPBYTE rgbSpriteRam1 = &RAM[0x8b80]; //	image and color
	LPBYTE rgbSpriteRam2 = &RAM[0x9380]; //	position
	LPBYTE rgbSpriteRam3 = &RAM[0x9b80]; // x/y flip, flags

	for (i = 0; i < 64; i++)
	{
		if (rgbSpriteRam3[2*i+1] & 2)
			continue;

		int x = rgbSpriteRam2[2*i];
		int y = rgbSpriteRam2[2*i+1] - 40 + 0x100 * (rgbSpriteRam3[2*i+1] & 1);
		int c = 128 + (rgbSpriteRam1[2*i+1] << 2);
		int n = (rgbSpriteRam1[2*i] & 0x7F) + 128 * (rgbSpriteRam3[2*i] & 3);

		if (rgbSpriteRam3[2*i] & 8) // double width (duel ship bullets)
		{
			if (x + 16 < 256)
			{
				if (y > 0 && y < 256 + 32)
				{
					_galaga.pimgSprites->DrawSprite(rgbBuffer, lPitch, x, y, n + 2, c);
					_galaga.pimgSprites->DrawSprite(rgbBuffer, lPitch, x + 16, y, n, c);
				}

				if (rgbSpriteRam3[2*i] & 4) // double height (explosions)
				{
					if (y > 0 && y < 256 + 16)
					{
						_galaga.pimgSprites->DrawSprite(rgbBuffer, lPitch, x, y + 16, n + 3, c);
						_galaga.pimgSprites->DrawSprite(rgbBuffer, lPitch, x + 16, y + 16, n + 1, c);
					}
				}
			}
		}
		else
		{				
			if (y > 0 && y < 256 + 32)
				_galaga.pimgSprites->DrawSprite(rgbBuffer, lPitch, x, y, n, c);
		}
	}

	//
	//  Draw 32x32 characters
	//

	for (i = 0; i < 32 * 32; i++)
	{
		const int offset_value = _galaga.pVideoRam[i];

		if (offset_value == 0x24)
			continue;
			
		const int color = _galaga.pColorRam[i] & 0x1F;

		int sx,sy,mx,my;

		mx = i / 32;
		my = i % 32;

#if 1
		if (mx <= 1)
		{
			sx = 29 - my;
			sy = mx + 34;
		}
		else if (mx >= 30)
		{
			sx = 29 - my;
			sy = mx - 30;
		}
		else
		{
			sx = 29 - mx;
			sy = my + 2;
		}
#else
		sx = mx;
		sy = my;
#endif

		const int x = 16 + 8 * sx;
		const int y = 8 * sy;

		// BUG:  DrawCharT here AVFs under NT!!!
		DrawChar(_galaga.pimgChars->rgbData, //_galaga.pimgChars->rgbMask,
			offset_value, color << 2, x, y, rgbBuffer, lPitch);
	}
}

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	// GalagaUpdateSound is in the sound module, with the rest of the mixing code
	GalagaUpdateSound(rgbBuffer, cbBuffer, &RAM[SOUND_BASE], _fSoundChanged);
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdGalaga = 
{
	TEXT("Galaga"),
	(LPVOID)0x00000000,
	GalagaInit,
	GalagaExec,
	GalagaRenderScene,
	UpdateSound,
	GalagaTerm
};

