/*---------------------------------------------------------------------------

	File:
		mariobro.cpp
	
	Abstract:
		Implements the "Mario Bros." game driver functions.

	Revisions:
		10-Feb-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Acknowledgements:
	
	Mario Bros. hardware information adopted from Mirko Buffoni's Mario Bros.
	driver for MAME.

	TODO:  add char colors (try making chars 3 or four bits, where the palette
		information is encoded in the character data itself!)

	TODO:  don't bother drawing the first and last two rows a chars!!

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

extern void SoundUpdateMixer(LPBYTE rgbBuffer, LONG cbBuffer);

static IMAGE *_pimgSprites;
static IMAGE *_pimgChars;
static IExecZ80 *_pZ80;

struct SPRITE
{
	BYTE y, c, i, x;
};

const SPRITE *rgSprites = (SPRITE*)&RAM[0x6900];

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

/*

sprite color schemes...

00 pipes
01 shell creeper (normal)
02 shell creeper ()
03 shell creeper
04 coin
05
06
07
08
09
0A mario
0B
0C
0D
0E platform	/ pow
0F pow
10
11
12
13
14
15
16
17
18
19
1A
1B
1C
1D
1E
1F
*/

static const PALETTEENTRY _rgpeMario[] =
{
	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   0,   255 },
	{ 150, 150, 0   },
	{ 200, 200, 0   },
	{ 255, 255, 255 },
	{ 0,   0,   0   },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 255, 255 },
	{ 150, 0,   0   },
	{ 0,   255, 0   }, // turtle shell
	{ 0,   0,   0   },
	{ 252, 200, 0   },
	{ 0,   0,   255 },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 255, 255 },
	{ 150, 0,   0   },
	{ 255, 0,   255 }, // turtle shell
	{ 0,   0,   0   },
	{ 252, 200, 0   },
	{ 0,   0,   255 },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 255, 255 },
	{ 150, 0,   0   },
	{ 255, 0,   0   }, // turtle shell
	{ 0,   0,   0   },
	{ 252, 200, 0   },
	{ 0,   0,   255 },

	{ 0,   0,   0   },
	{ 155, 0,   0   },
	{ 255, 255, 0   },
	{ 64,  64,  64  },
	{ 255, 255, 255 },
	{ 150, 150, 0   },
	{ 255, 0,   0   }, // side stepper shell
	{ 0,   0,   150 },

	{ 0,   0,   0   },
	{ 155, 0,   0   },
	{ 255, 255, 0   },
	{ 64,  64,  64  },
	{ 255, 255, 255 },
//	{ 150, 150, 0   },
	{ 0,   0,   255 },
	{ 255, 0,   255 }, // side stepper shell
	{ 0,   0,   150 },

	{ 0,   0,   0   },
	{ 155, 0,   0   },
	{ 255, 255, 0   },
	{ 64,  64,  64  },
	{ 255, 255, 255 },
	{ 150, 150, 0   },
	{ 0,   255, 0   }, // side stepper shell
	{ 0,   0,   150 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 255, 0   },
	{ 0,   0,   255 },
	{ 0,   255, 255 },
	{ 0,   255, 255 },
	{ 255, 0,   255 },
	{ 0,   0,   200 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 0,   0   },
	{ 150, 150, 0   },
	{ 0,   0,   216 },
	{ 252, 0,   0   },
	{ 252, 184, 184 },
	{ 0,   0,   0   },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 100, 100, 0   },
	{ 255, 0,   255 },
	{ 0,   216, 0   },
	{ 252, 0,   0   },
	{ 252, 184, 184 },
	{ 0,   0,   0   },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	{ 0,   0,   0   },
	{ 255, 0,   0   },
	{ 255, 255, 0   },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 0,   255 },
	{ 255, 255, 255 },

	// POW, platform
	
	{ 0,   0,   0   }, 
	{ 0,   0,   0   },
	{ 0,   0,   255 },
	{ 0,   255, 0   },
	{ 0,   255, 255 },
	{ 0,   0,   255 },
	{ 255, 255, 255 },
	{ 255, 255, 255 },

	// POW
	
	{ 0,   0,   0   }, 
	{ 0,   0,   0   },
	{ 155, 155, 0   },
	{ 0,   255, 0   },
	{ 255, 255, 0   },
	{ 0,   0,   255 },
	{ 255, 255, 255 },
	{ 255, 255, 255 },  

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 255, 0,   0   },
	{ 150, 150, 0   },
	{ 0,   0,   216 },
	{ 252, 0,   0   },
	{ 252, 184, 184 },
	{ 0,   0,   0   },

	{ 0,   0,   0   },
	{ 255, 255, 255 },
	{ 100, 100, 0   },
	{ 255, 0,   255 },
	{ 0,   216, 0   },
	{ 252, 0,   0   },
	{ 252, 184, 184 },
	{ 0,   0,   0   },

	// text!!!
	
	{ 0,   0,   0   },
	{ 0,   0,   255 },
	{ 255, 255, 255 },
	{ 0,   0,   255 },
	{ 150, 150, 0   },
	{ 200, 200, 0   },
	{ 255, 255, 255 },
	{ 0,   0,   0   },
};

static const int _rgiCharColors[128][4] = {
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },

	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
	{ 0, 1, 2, 3 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0x7C00:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x20;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x40;

			if (osd_key_pressed(OSD_KEY_F1))
				b ^= 0x80;

			return b;
		}

		case 0x7C80:
		{
			BYTE b = 0x00;

			if (osd_key_pressed(OSD_KEY_X))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_Z))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_SPACE))
				b ^= 0x10;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x20;

			return b;
		}

		default:
			return RAM[A];
	}
}

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x4000)
		return;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const TCHAR _szMbPath[] = "mario";

RAGE_ROMTABLE_BEGIN(MbROMs)
	RAGE_ROMTABLE_ENTRY("mario.7f", 0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("mario.7e", 0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("mario.7d", 0x4000, 0x2000)
	RAGE_ROMTABLE_ENTRY("mario.7c", 0xF000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(MbGFXROMs)
	RAGE_ROMTABLE_ENTRY("mario.3f", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.3j", 0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7m", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7n", 0x3000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7p", 0x4000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7s", 0x5000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7t", 0x6000, 0x1000)
	RAGE_ROMTABLE_ENTRY("mario.7u", 0x7000, 0x1000)
RAGE_ROMTABLE_END

static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
static const RECT _rcDisplay = { 0, 16 + 8, 256, 240 + 8};

static const DECODE_GRAPHICS dgsChars =
{
	0x00000000,
	{
		{ 8, 8 },
		512,
		8 * 8,
		2,
		{ 0, 8 * 8 * 512 },
		2,
		{ 7, 6, 5, 4, 3, 2, 1 },
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

static const DECODE_GRAPHICS dgs =
{
	0x00000001,
	{
		{ 16, 16 },
		256,
		16 * 8,
		3,
		{ 0, 16 * 16 * 256, 2 * 16 * 16 * 256 },
		3,
		{ 256*16*8+7, 256*16*8+6, 256*16*8+5, 256*16*8+4, 256*16*8+3, 256*16*8+2, 256*16*8+1, 256*16*8+0,
			7, 6, 5, 4, 3, 2, 1, 0 },	/* the two halves of the sprite are separated */
		{ 15*8, 14*8, 13*8, 12*8, 11*8, 10*8, 9*8, 8*8,
			7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
	},

	1,
	
	4, 
	
	{
		0 * 256,
		2 * 256,
		1 * 256,
		3 * 256
	},
	0, 
	NULL
};

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
//		{ 0x0000, 0xFFFF, 0, NULL, NULL },
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
/*
	{ 0x7C00, 0x7C00, 0, 
	{ 0x7C80, 0x7C80, 0, M_RDMEM, M_WRMEM },
	{ 0x7D00, 0x7D07, 0, M_RDMEM, M_WRMEM },
	{ 0x7D80, 0x7D80, 0, M_RDMEM, M_WRMEM },
*/
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	int i, j, k;

	for (i = 0; i < DIM(_rgpeMario); i++)
		gli.ppe[i] = _rgpeMario[i];

	gli.cpe = i;

	if (readroms(RAM, MbROMs, _szMbPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//

	BYTE tmpstorage[0x8000];

	if (readroms(tmpstorage, MbGFXROMs, _szMbPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//  Decode characters
	//

	MyGfxDecode3(dgsChars, tmpstorage, _pimgChars);

	for (i = 0; i < dgsChars.mgd.cImage; i++)
	{
		for (j = 0; j < dgsChars.mgd.sizImage.cy; j++)
		{
			for (k = 0; k < dgsChars.mgd.sizImage.cx; k++)
			{
				// NOTE:  assumes char data was decoded in LOWER bits of _rgbChars...
				_pimgChars->rgbData[i * 64 + j * 8 + k] = _rgiCharColors[i >> 2][_pimgChars->rgbData[i * 64 + j * 8 + k]];
			}
		}
	}

	//
	//  Decode sprites
	//
	
	MyGfxDecode3(dgs, tmpstorage + 0x2000, _pimgSprites);

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pfnMemRead = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _pZ80);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

static void Term()
{
	_pZ80->Release();

	FreeImage(_pimgChars);
	FreeImage(_pimgSprites);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	SoundUpdateMixer(rgbBuffer, cbBuffer);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int Exec()
{
	int cCycles;
	
	cCycles = _pZ80->Exec(50000);
	
	if (RAM[0x7e84] != 0)
		cCycles += _pZ80->Nmi();

	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const int _rgiHex[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 17, 18, 19, 20, 21, 22 };

static _iColor = 9;

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;

#if 0
	if (GetAsyncKeyState('C') < 0)
	{
		for (i = 0; i < 512; i++)
			DrawChar((LPBYTE)_rgbChars, i, _iColor << 3, 8 * (i % 32), 8 * (i / 32), rgbBuffer, lPitch);

		return S_OK;
	}

	if (GetAsyncKeyState('S') < 0)
	{
		for (i = 0; i < 32 * 32; i++)
			DrawChar((LPBYTE)_rgbChars, 0x50, _iColor << 3, 8 * (i % 32), 8 + 8 * (i / 32), rgbBuffer, lPitch);

		for (i = 0; i < 256; i++)
			DrawSprite(rgbBuffer, lPitch, 16 * (i % 16), 16 * (i / 16), i, _iColor << 3);

		return S_OK;
	}

	const BOOL fShowI = GetAsyncKeyState('I') < 0;
	const BOOL fShow0 = GetAsyncKeyState('0') < 0;
#endif

	//
	//  Draw 32x32 characters
	//

	for (i = 0; i < 32 * 32; i++)
	{
		const int x = 8 * (31 - (i % 32));
		const int y = 8 + 8 * (31 - (i / 32));

		DrawChar(_pimgChars->rgbData, RAM[0x7400 + i] + 256 * RAM[0x7e80],
			x, y, rgbBuffer, lPitch);
	}

	//
	//  Draw sprites
	//

	for (i = 0; i < 96; i++)
	{
		if (rgSprites[i].y == 0)
			continue;

#if 0
		if (fShowI)
		{
			DrawChar(rgbBuffer, lPitch, 256 - ps->x - 8, ps->y,
				((ps->c & 0x3F) & 0xF0) >> 4, 0);

			DrawChar(rgbBuffer, lPitch, 256 - ps->x, ps->y,
				((ps->c & 0x3F) & 0x0F), 0);
		}
		else
		{
			if (fShow0)
			{
				DrawSprite(rgbBuffer, lPitch, 256 - ps->x - 8, ps->y,
					ps->i + ((ps->c & 0x80) << 1) + ((ps->c & 0x40) << 3), _iColor << 3);
			}
			else
			{
				DrawSprite(rgbBuffer, lPitch, 256 - ps->x - 8, ps->y,
					ps->i + ((ps->c & 0x80) << 1) + ((ps->c & 0x40) << 3), (ps->c & 0x3F) << 3);
			}
		}
#else
		DrawSprite(_pimgSprites->rgbData, _pimgSprites->rgbMask,
			rgSprites[i].i + ((rgSprites[i].c & 0x80) << 1) + ((rgSprites[i].c & 0x40) << 3),
			(rgSprites[i].c & 0x3F) << 3, 
			256 - rgSprites[i].x - 8, rgSprites[i].y,
			rgbBuffer, lPitch);
#endif
	}
}

/*---------------------------------------------------------------------------

	Define game descriptors

---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdMarioBros = 
{
	TEXT("Mario Bros."),
	NULL,
	Init,
	Exec,
	RenderScene,
	UpdateSound,
	Term
};

