/*---------------------------------------------------------------------------

	File:
		mrdo.cpp
	
	Abstract:
		Implements the "Mr. Do!" game driver functions.

	Revisions:
		09-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information adopted from Nicola Salmoria's MAME.

	TODO:  decode color PROM and build composite palette
	TODO:  implement backmost character layer
	TODO:  decode chars properly

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"
#include <math.h>

// note: assumes byte packing...

struct SPRITE
{
	BYTE i, x, c, y;
};

struct MRDO_GAME
{
	SPRITE *pSprites;
	BYTE *pVideoRam1;
	BYTE *pColorRam1;
	BYTE *pVideoRam2;
	BYTE *pColorRam2;
	BYTE *pXScroll;

	DECODE_GRAPHICS dgChars1;
	DECODE_GRAPHICS dgChars2;
	DECODE_GRAPHICS dgSprites;
	IMAGE *pimgChars1;
	IMAGE *pimgChars2;
	IMAGE *pimgSprites;
	IExecZ80 *pZ80;
	SIZE sizBuffer;
	RECT rcBuffer;
	BYTE **ppSoundSamples;
	BYTE *pColorProm;
	RomModule rgRomModuleMain[16];
	RomModule *prmChars;
	RomModule *prmSprites;

	static void RenderScene(LPBYTE rgbBuffer, LONG lPitch);
	static BYTE M_RDMEM(WORD A);
	static void M_WRMEM(WORD A, BYTE V);
	static HRESULT Init(GAME_LOAD_INFO &gli);

	TCHAR szTitle[64];
	static const TCHAR szPath[64];

	TCHAR rgszRomModuleMain[16][64];

	static const BYTE _rgbColorPROM[];
	static const DECODE_GRAPHICS dgsChars1, dgsChars2, dgsSprites;
};

static BYTE _rgbToneSample[2000];
static BYTE _rgbNoiseSample[10000];

#define TONE_PERIOD 4
#define NOISE_LENGTH 10000
#define WAVE_AMPLITUDE 70
#define SND_CLOCK 4000000	/* 3.072 Mhz */

static MRDO_GAME _game =
{
	(SPRITE*)&RAM[0x9000],
	&RAM[0x8400],
	&RAM[0x8000],
	&RAM[0x8C00],
	&RAM[0x8800],
	&RAM[0xF800],
};

struct SN76496
{
	/* set this before calling SN76496Reset() */
	int Clock;	/* chip clock in Hz */

	/* read these after calling SN76496Update(), and produce the required sounds */
	int Volume[4];	/* volume of voice 0-2 and noise. Range is 0-255 */
	int Frequency[3];	/* tone frequency in Hz */
	int NoiseShiftRate;	/* sample rate for the noise sample */

	/* private */
	int Register[8];
	int LastRegister;
};

void SN76496Reset(struct SN76496 *R)
{
	int i;


	for (i = 0;i < 4;i++) R->Volume[i] = 0;
	for (i = 0;i < 3;i++) R->Frequency[i] = 1000;
	R->NoiseShiftRate = R->Clock/64;

	R->LastRegister = 0;
	for (i = 0;i < 8;i+=2)
	{
		R->Register[i] = 0;
		R->Register[i + 1] = 0x0f;	/* volume = 0 */
	}
}



void SN76496Write(struct SN76496 *R,int data)
{
	if (data & 0x80)
	{
		R->LastRegister = (data >> 4) & 0x07;
		R->Register[R->LastRegister] = data & 0x0f;
	}
	else
	{
		R->Register[R->LastRegister] =
				(R->Register[R->LastRegister] & 0x0f) | ((data & 0x3f) << 4);
	}
}



void SN76496Update(struct SN76496 *R)
{
	int i,n;


	for (i = 0;i < 4;i++)
	{
		n = 0x0f - R->Register[2 * i + 1];

		R->Volume[i] = (n << 4) | n;
	}

	for (i = 0;i < 3;i++)
	{
		n = R->Register[2 * i];

		/* if period is 0, shut down the voice */
		if (n == 0) R->Volume[i] = 0;
		else R->Frequency[i] = R->Clock / (32 * n);
	}

	R->NoiseShiftRate = R->Clock/64;

	n = R->Register[6] & 3;
	if (n == 3)
	{
		static int count;


		/* kludge just to play something */
		count++;
		if (count > 2) count = 0;
		R->NoiseShiftRate = 64 << count;
	}
	else R->NoiseShiftRate = 64 << n;
}

#define CHIPS 2


#define TONE_LENGTH 2000
#define TONE_PERIOD 4
#define NOISE_LENGTH 10000
#define WAVE_AMPLITUDE 70


static char *tone;
static char *noise;

static struct SN76496 sn[CHIPS];

/*---------------------------------------------------------------------------

	Static data

---------------------------------------------------------------------------*/

const TCHAR MRDO_GAME::szPath[] = "mrdo";

static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
static const RECT _rcDisplay = { 32, 16, 256 - 32, 256 + 16};

RAGE_ROMTABLE_BEGIN(ROMs)
	RAGE_ROMTABLE_ENTRY("D1", 0x0000, 0x2000)
	RAGE_ROMTABLE_ENTRY("D2", 0x2000, 0x2000)
	RAGE_ROMTABLE_ENTRY("D3", 0x4000, 0x2000)
	RAGE_ROMTABLE_ENTRY("D4", 0x6000, 0x2000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(GFXROMs)
	RAGE_ROMTABLE_ENTRY("D9", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("D10",0x1000, 0x1000)
	RAGE_ROMTABLE_ENTRY("D8", 0x2000, 0x1000)
	RAGE_ROMTABLE_ENTRY("D7", 0x3000, 0x1000)
	RAGE_ROMTABLE_ENTRY("D5", 0x4000, 0x1000)
	RAGE_ROMTABLE_ENTRY("D6", 0x5000, 0x1000)
RAGE_ROMTABLE_END

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
	{ 0x0000, 0xFFFF, 0, MRDO_GAME::M_RDMEM, MRDO_GAME::M_WRMEM },
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

const BYTE MRDO_GAME::_rgbColorPROM[] =
{
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
	0x00, 0xFF, 0x07, 0x38,
};

static const BYTE color_prom[] =
{
	/* palette (high bits) */
	0x00,0x0C,0x03,0x00,0x0F,0x0B,0x0C,0x3F,0x0D,0x0F,0x0F,0x0C,0x0C,0x3C,0x0C,0x30,
	0x0C,0x03,0x30,0x03,0x0C,0x0F,0x00,0x3F,0x03,0x1E,0x00,0x0F,0x37,0x36,0x0D,0x33,
	/* palette (low bits) */
	0x00,0x0C,0x03,0x00,0x0C,0x03,0x00,0x3F,0x0F,0x03,0x0F,0x3F,0x0C,0x0F,0x0F,0x3A,
	0x03,0x0F,0x00,0x0C,0x00,0x0F,0x3F,0x03,0x2A,0x0C,0x00,0x0A,0x0C,0x0E,0x3F,0x0F,
	/* sprite color lookup table */
	0x00,0x97,0x71,0xF9,0x00,0x27,0xA5,0x13,0x00,0x32,0x77,0x3F,0x00,0xA7,0x72,0xF9,
	0x00,0x1F,0x9A,0x77,0x00,0x15,0x27,0x38,0x00,0xC2,0x55,0x69,0x00,0x7F,0x76,0x7A
};

const DECODE_GRAPHICS MRDO_GAME::dgsChars1 =
{
	0x00000001,
	{
		{ 8, 8 },
		512,
		8 * 8,
		2,
		{ 0, 8 * 8 * 512 },
		2,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

const DECODE_GRAPHICS MRDO_GAME::dgsChars2 =
{
	0x00000001,
	{
		{ 8, 8 },
		512,
		8 * 8,
		2,
		{ 0, 8 * 8 * 512 },
		2,
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
	},

	1,
	
	1, 
	
	{
		0
	},
	0, 
	NULL
};

const DECODE_GRAPHICS MRDO_GAME::dgsSprites =
{
	0x00000001,
	{
		{ 16, 16 },
		128,
		64 * 8,
		2,
		{ 0, 4 },
		2,
		{ 0*16, 2*16, 4*16, 6*16, 8*16, 10*16, 12*16, 14*16,
			16*16, 18*16, 20*16, 22*16, 24*16, 26*16, 28*16, 30*16 },
		{ 24+0, 24+1, 24+2, 24+3, 16+0, 16+1, 16+2, 16+3,
			8+0, 8+1, 8+2, 8+3, 0, 1, 2, 3 },
	},

	1,
	
	4, 
	
	{
		0 * 128,
		1 * 128,
		2 * 128,
		3 * 128
	},
	0, 
	NULL
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void BuildCompositePalette(PALETTEENTRY *ppeOut, UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < DIM(MRDO_GAME::_rgbColorPROM); cpeOut++)
	{
		int i = cpeOut;

		int bit0,bit1,bit2;

		bit0 = (MRDO_GAME::_rgbColorPROM[i] >> 0) & 0x01;
		bit1 = (MRDO_GAME::_rgbColorPROM[i] >> 1) & 0x01;
		bit2 = (MRDO_GAME::_rgbColorPROM[i] >> 2) & 0x01;

		ppeOut[cpeOut].peRed = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = (MRDO_GAME::_rgbColorPROM[i] >> 3) & 0x01;
		bit1 = (MRDO_GAME::_rgbColorPROM[i] >> 4) & 0x01;
		bit2 = (MRDO_GAME::_rgbColorPROM[i] >> 5) & 0x01;

		ppeOut[cpeOut].peGreen = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

		bit0 = 0;
		bit1 = (MRDO_GAME::_rgbColorPROM[i] >> 6) & 0x01;
		bit2 = (MRDO_GAME::_rgbColorPROM[i] >> 7) & 0x01;

		ppeOut[cpeOut].peBlue  = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
	}
}

HRESULT MRDO_GAME::Init(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(gli.ppe, gli.cpe);

	//
	//
	//
	
	if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	BYTE tmpstorage[0x6000];

	if (readroms(tmpstorage, GFXROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//  Decode graphics
	//

	MyGfxDecode3(dgsChars1, tmpstorage, _game.pimgChars1);
	MyGfxDecode3(dgsChars2, tmpstorage + 0x2000, _game.pimgChars2);
	MyGfxDecode3(dgsSprites, tmpstorage + 0x4000, _game.pimgSprites);

	//
	//
	//

	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pfnMemRead = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd = rgmd;

	HRESULT hr = EclCreateZ80(cs, _game.pZ80);

	if (FAILED(hr))
		return hr;

	//
	//
	//

	#define PI 3.1415

	for (int i = 0;i < DIM(_rgbToneSample);i++)
		_rgbToneSample[i] = BYTE(WAVE_AMPLITUDE * sin(2*PI*i/TONE_PERIOD));

	for (i = 0;i < DIM(_rgbNoiseSample);i++)
		_rgbNoiseSample[i] = (rand() % (2*WAVE_AMPLITUDE)) - WAVE_AMPLITUDE;

	for (int j = 0;j < CHIPS;j++)
	{
		sn[j].Clock = SND_CLOCK;

		SN76496Reset(&sn[j]);

/*
		for (i = 0;i < 3;i++)
			osd_play_sample(4*j+i,tone,TONE_LENGTH,TONE_PERIOD * sn[j].Frequency[i],sn[j].Volume[i],1);

		osd_play_sample(4*j+3,noise,NOISE_LENGTH,sn[j].NoiseShiftRate,sn[j].Volume[3],1);
*/
	}

	return S_OK;
}

static void Term()
{
	_game.pZ80->Release();

	FreeImage(_game.pimgChars1);
	FreeImage(_game.pimgChars2);
	FreeImage(_game.pimgSprites);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

BYTE MRDO_GAME::M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0xA000:
		{
			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x08;

			if (osd_key_pressed(OSD_KEY_CONTROL))
				b ^= 0x10;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x20;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x40;

			return b;
		}

		case 0xA001:
		{
			BYTE b = 0xFF;

			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x40;

			return b;
		}

		case 0xA002:
		{
			BYTE b = 0xDF;

			if (osd_key_pressed(OSD_KEY_F1))
				b ^= 0x04;

			return b;
		}

		case 0xA003:
		{
			BYTE b = 0xFF;

			return b;
		}

		case 0x9803:
		{
			EXECZ80_CONTEXT ctx;
			
			_game.pZ80->GetContext(ctx);
			
			return RAM[ctx.HL];
		}
		 
		default:
			return RAM[A];
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void MRDO_GAME::M_WRMEM(WORD A, BYTE V)
{
	if (A < 0x8000)
		return;

	switch (A)
	{
		case 0x9801:
		{
			SN76496Write(&sn[0],V);
		}

		case 0x9802:
		{
			SN76496Write(&sn[1],V);
		}
	}

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int Exec()
{
	int cCycles;
	
	cCycles = _game.pZ80->Exec(66000);
	cCycles += _game.pZ80->Irq(0xFF);

	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void MRDO_GAME::RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

#if 0
	for (i = 0; i < 32; i++)
	{
		for (j = 4; j < 28; j++)
		{
			DrawChar(rgbBuffer, lPitch, 8 * (31 - j), 8 * (31 - i),
				RAM[0x8400 + i + j * 32], RAM[0x8000 + i + j * 32] & 0x0F);
		}
	}
#else
/*
	for (i = 0; i < 32; i++)
	{
		for (j = 0; j < 32; j++)
		{
			DrawChar(_game.pimgChars2->rgbData, RAM[0x8400 + i + j * 32] + 2 * (RAM[0x8000 + i + j * 32] & 0x80),
				RAM[0x8000 + i + j * 32] & 0x7F, 8 * j, 16 + 8 * (31 - i), rgbBuffer, lPitch);
		}
	}

	for (i = 0; i < 32; i++)
	{
		for (j = 4; j < 28; j++)
		{
			DrawCharT(_game.pimgChars1->rgbData, _game.pimgChars1->rgbMask, RAM[0x8C00 + i + j * 32] + 2 * (RAM[0x8800 + i + j * 32] & 0x80),
				RAM[0x8800 + i + j * 32] & 0x7F, 8 * j, 16 + 8 * (31 - i), rgbBuffer, lPitch);
		}
	}
*/

	for (i = 0; i < 32; i++)
	{
		for (j = 4; j < 28; j++)
		{
			DrawChar(_game.pimgChars1->rgbData, RAM[0x8C00 + i + j * 32] + 2 * (RAM[0x8800 + i + j * 32] & 0x80),
				RAM[0x8800 + i + j * 32] & 0x7F, 8 * j, 16 + 8 * (31 - i), rgbBuffer, lPitch);
		}
	}
#endif

	const SPRITE *rgs = (SPRITE*)&RAM[0x9000];

	for (i = 63; i >= 0; i--)
	{
		if (rgs[i].x != 0)
		{
			DrawSprite(_game.pimgSprites->rgbData, _game.pimgSprites->rgbMask,
				(rgs[i].i & 0x7F) + ((rgs[i].c & 0x30) * 8), (rgs[i].c & 0x0F) << 2,
				255 - rgs[i].x, 255 - rgs[i].y, rgbBuffer, lPitch);
		}
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void MrdoUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	for (int j = 0;j < CHIPS;j++)
		SN76496Update(&sn[j]);

/*
		for (i = 0;i < 3;i++)
			osd_adjust_sample(4*j+i,TONE_PERIOD * sn[j].Frequency[i],sn[j].Volume[i]);

		osd_adjust_sample(4*j+3,sn[j].NoiseShiftRate,sn[j].Volume[3]);
*/

	//
	//
	//

	static int rgiPos[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
	
	while (cbBuffer--)
	{
		//
		//  initialize accumulator (destination buffer is unsigned)
		//
		
		int nAcc = 0;

		//
		//  loop through each channel
		//

		for (int j = 0;j < CHIPS;j++)
		{
			for (int iv = 0; iv < 3; iv++)
			{
				nAcc += _rgbToneSample[rgiPos[iv + 4 * j] >> 12] * sn[j].Volume[iv];

				rgiPos[iv + 4 * j] += (TONE_PERIOD * sn[j].Frequency[iv] / 10);
//				rgiPos[iv + 4 * j] += (1 << 12) * DIM(_rgbToneSample) * sn[j].Frequency[iv] / 11025;

				if (rgiPos[iv + 4 * j] > (DIM(_rgbToneSample) << 12))
					rgiPos[iv + 4 * j] -= (DIM(_rgbToneSample) << 12);
			}

			nAcc += _rgbNoiseSample[rgiPos[3 + 4 * j] >> 12] * sn[j].Volume[3];

			rgiPos[3 + 4 * j] += sn[j].NoiseShiftRate;

			if (rgiPos[3 + 4 * j] > (DIM(_rgbToneSample) << 12))
				rgiPos[3 + 4 * j] -= (DIM(_rgbToneSample) << 12);
		}

		//
		//  clip...
		//
		
		nAcc /= 256 * 8;
		
		nAcc += 128;

		if (nAcc < 0)
			nAcc = 0;
		else if (nAcc > 255)
			nAcc = 255;

		//
		//
		//

		*rgbBuffer++ = nAcc;
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdMrDo = 
{
	TEXT("Mr. Do!"),
	NULL,
	MRDO_GAME::Init,
	Exec,
	MRDO_GAME::RenderScene,
	MrdoUpdateSound,
	Term
};

