/*---------------------------------------------------------------------------

	File:
		mrdo.h
	
	Abstract:
		Implements the Mr. Do! game driver functions.

	Revisions:
		05-Feb-97 EAM Created

Mr Do! memory map (preliminary)

0000-7fff ROM
8000-83ff color RAM 1
8400-87ff video RAM 1
8800-8bff color RAM 2
8c00-8fff video RAM 2
e000-efff RAM

memory mapped ports:

read:
9803      SECRE 1/6-J2-11
a000      IN0
a001      IN1
a002      DSW1
a003      DSW2

write:
9000-90ff sprites, 64 groups of 4 bytes.
9800      flip (bit 0) playfield priority selector? (bits 1-3)
9801      sound port 1
9802      sound port 2
f000      playfield 0 Y scroll position (not used by Mr. Do!)
f800      playfield 0 X scroll position

---------------------------------------------------------------------------*/


#define TOTAL_COLORS 9
#define COLOR_CODES 24

/* color codes used to print text on the screen */
#define WHITE_TEXT 0x03
#define YELLOW_TEXT 0x02
#define RED_TEXT 0x00
/* start of numbers and letters in the character roms */
#define NUMBERS_START 0
#define LETTERS_START 10

/* CPU_SPEED is the speed of the CPU in Hz. It is used together with */
/* FRAMES_PER_SECOND to calculate how much CPU cycles must pass between */
/* interrupts. */
#define CPU_SPEED 4000000	/* 4 Mhz */
#define FRAMES_PER_SECOND 60

/* I'm not yet sure about how the vertical blanking should be handled. */
/* I think that IN1_VBLANK should be 1 during the whole vblank, which */
/* should last roughly 1/12th of the frame. */
#define VBLANK_CYCLES ((CPU_SPEED / FRAMES_PER_SECOND) / 12)


enum {BLACK,WHITE,RED,GREEN,BLUE,ORANGE,YELLOW,PURPLE,GRAY};
#define WALLS PURPLE


#define SND_CLOCK 4000000	/* 4 Mhz */


/* Lady Bug has a 24x30 characters screen, but the video memory is 32x32 */
#define H_CHARS	32
#define V_CHARS 32
#define BITMAP_WIDTH (8*H_CHARS)
#define BITMAP_HEIGHT (8*V_CHARS)

#define FIRST_VISIBLE_ROW 1
#define LAST_VISIBLE_ROW (V_CHARS-1)
#define FIRST_VISIBLE_COLUMN 4
#define LAST_VISIBLE_COLUMN (H_CHARS-4)

#define TOTAL_CHARS 1024
#define TOTAL_SPRITES 128
#define SPRITE_WIDTH 16
#define SPRITE_HEIGHT 16

/////////////////////////////////////////////////////////////////


#define ROM_END 0x7fff

#define VIDEO1_RAM_START 0x8400
#define VIDEO2_RAM_START 0x8c00
#define VIDEO_RAM_SIZE 0x400
#define COLOR1_RAM_START (VIDEO1_RAM_START-VIDEO_RAM_SIZE)
#define COLOR2_RAM_START (VIDEO2_RAM_START-VIDEO_RAM_SIZE)

#define SCROLL_X 0xf800
#define SCROLL_Y 0xf000

#define SPRITE_BASE 0x9000

#define SOUND_PORT_1 0x9801
#define SOUND_PORT_2 0x9802


/*
 * IN0 (all bits are inverted)
 * bit 7 : TILT if this is 0 coins are not accepted
 * bit 6 : START 2
 * bit 5 : START 1
 * bit 4 : FIRE player 1
 * bit 3 : UP player 1
 * bit 2 : RIGHT player 1
 * bit 1 : DOWN player 1
 * bit 0 : LEFT player 1
 */
#define IN0_PORT 0xa000
#define IN0_TILT (1<<7)
#define IN0_START2 (1<<6)
#define IN0_START1 (1<<5)
#define IN0_FIRE (1<<4)
#define IN0_UP (1<<3)
#define IN0_RIGHT (1<<2)
#define IN0_DOWN (1<<1)
#define IN0_LEFT (1<<0)

/*
 * IN1 (player input bits are inverted)
 * bit 7 : COIN 2
 * bit 6 : COIN 1
 * bit 5 : unused
 * bit 4 : FIRE player 2 (TABLE only)
 * bit 3 : UP player 2 (TABLE only)
 * bit 2 : RIGHT player 2 (TABLE only)
 * bit 1 : DOWN player 2 (TABLE only)
 * bit 0 : LEFT player 2 (TABLE only)
 */
#define IN1_PORT 0xa001
#define IN1_COIN2 (1<<7)
#define IN1_COIN1 (1<<6)
#define IN1_UNUSED (1<<5)
#define IN1_FIRE (1<<4)
#define IN1_UP (1<<3)
#define IN1_RIGHT (1<<2)
#define IN1_DOWN (1<<1)
#define IN1_LEFT (1<<0)

/*
 * DSW1 (all bits are inverted)
 * bit 7 : DIP SWITCH 8\ Number of lives
 * bit 6 : DIP SWITCH 7/ 11 = 3  10 = 4  01 = 5  11 = 6
 * bit 5 : DIP SWITCH 6  TABLE or STANDUP (0 = STANDUP)
 * bit 4 : DIP SWITCH 5  "EXTRA" difficulty  1 = easy  0 = hard
 * bit 3 : DIP SWITCH 4  "SPECIAL" difficulty  1 = easy  0 = hard
 * bit 2 : DIP SWITCH 3  RACK TEST
 * bit 1 : DIP SWITCH 2\ Difficulty level
 * bit 0 : DIP SWITCH 1/ 11 = 1st (easy) 10 = 2nd 01 = 3rd 00 = 4th (hard)
 */
#define DSW1_PORT 0xa002
#define DSW1_SW8 (1<<7)
#define DSW1_SW7 (1<<6)
#define DSW1_SW6 (1<<5)
#define DSW1_SW5 (1<<4)
#define DSW1_SW4 (1<<3)
#define DSW1_SW3 (1<<2)
#define DSW1_SW2 (1<<1)
#define DSW1_SW1 (1<<0)

#define DSW1_RACK_TEST (DSW1_SW3)

/*
 * DSW2 (all bits are inverted)
 * bit 7 : DIP SWITCH 8\
 * bit 6 : DIP SWITCH 7| Left coin slot
 * bit 5 : DIP SWITCH 6|
 * bit 4 : DIP SWITCH 5/
 * bit 3 : DIP SWITCH 4\
 * bit 2 : DIP SWITCH 3| Right coin slot
 * bit 1 : DIP SWITCH 2|
 * bit 0 : DIP SWITCH 1/
 *                       1111 = 1 coin 1 play
 *                       1110 = 1 coin 2 plays
 *                       1101 = 1 coin 3 plays
 *                       1100 = 1 coin 4 plays
 *                       1011 = 1 coin 5 plays
 *                       1010 = 2 coins 1 play
 *                       1001 = 2 coins 3 plays
 *                       1000 = 3 coins 1 play
 *                       0111 = 3 coins 2 plays
 *                       0110 = 4 coins 1 play
 *                       0000 = FREE PLAY
 *                 all others = 1 coin 1 play
 */
#define DSW2_PORT 0xa003
