/*---------------------------------------------------------------------------
	
	File:
		phoenix.cpp
	
	Abstract:
		Implements the Phoenix game driver functions.  Handles all the
		game's hardware emulation except CPU, loading ROMs.

	Revisions:
		04-Feb-97 EAM Created (from Robert's and Richard's source)
		05-Mar-97 EAM Fixed charset B decoding bug

	Copyright (C) 1997 Edward Massey

	Many thanks to Richard Davies for his Phoenix emulator source code
	(available at the Repository).

	Phoenix sound hardware information by ???.

	CONSIDER:  Compile the palette into the charset? (1024 images --> 64K)
	CONSIDER:  Combine drawing both charsets into one operation.  So that
	           video memory access is sequential and minimized.
	CONSIDER:  move drawing routines into blit
	CONSIDER:  store palette in a resource
	CONSIDER:  store sound waves in a resource
	CONSIDER:  read ROMs from script

	TODO:  Clean up scrolling background code

	Yet more speed ups...
	 - create 128 entry color palette based on color table
	 - put color information (offset_value >> 5) directly in char data
	 - repeat the 512 chars by using the second palette
	 - interleave the sprite data and mask

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "phoenix.h"
#include "game.h"
#include "game1.h"
#include "../z80/execz80.h"

struct MACHINE_STATE
{
	int interrupt_set;
	/* use these to draw charset B */
	unsigned char scrollreg;
	/* use this to select palette */
	unsigned char palreg;
	/* may use this to select video bank - is this ever used?*/
	unsigned char bankreg;

	IMAGE *pimgChars1;
	IMAGE *pimgChars2;
	IExecZ80 *pZ80;
};

static MACHINE_STATE _ms;

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

//
//
//

static const TCHAR szPath[] = "phoenix";

RAGE_ROMTABLE_BEGIN(ROMs)
	RAGE_ROMTABLE_ENTRY("phoenix.45", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.46", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.47", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.48", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.49", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.50", 0x2800, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.51", 0x3000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.52", 0x3800, 0x0800)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(GFXROMs)
	RAGE_ROMTABLE_ENTRY("phoenix.39", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.40", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.23", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("phoenix.24", 0x1800, 0x0800)
RAGE_ROMTABLE_END

//
//
//

static const TCHAR szPleiadesPath[] = "pleiads";

RAGE_ROMTABLE_BEGIN(PleiadesGFXROMs)
	RAGE_ROMTABLE_ENTRY("pleiades.27", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.26", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.45", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.44", 0x1800, 0x0800)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(PleiadesROMs)
	RAGE_ROMTABLE_ENTRY("pleiades.47", 0x0000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.48", 0x0800, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.49", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.50", 0x1800, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.51", 0x2000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.52", 0x2800, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.53", 0x3000, 0x0800)
	RAGE_ROMTABLE_ENTRY("pleiades.54", 0x3800, 0x0800)
RAGE_ROMTABLE_END

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

//
//
//

static const int base = 0x4000;
static const int MAXCHARS = 512;
static const int MAXSPRITES = 128;
static const int MAXPALETTES = 16;

/* CPU_SPEED is the speed of the CPU in Hz. It is used together with */
/* FRAMES_PER_SECOND to calculate how much CPU cycles must pass between */
/* interrupts. */
#define CPU_SPEED 3000000       /* 3 Mhz */
#define FRAMES_PER_SECOND 60

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const PALETTEENTRY _rgpe[] =
{
	{ 0x00, 0x00, 0x00 }, // BLACK
	{ 0xdb, 0xdb, 0xdb }, // WHITE
	{ 0xff, 0x00, 0x00 }, // RED
	{ 0x00, 0xff, 0x00 }, // GREEN
	{ 0x24, 0x24, 0xdb }, // BLUE
	{ 0x00, 0xff, 0xdb }, // CYAN,
	{ 0xff, 0xff, 0x00 }, // YELLOW,
	{ 0xff, 0xb6, 0xdb }, // PINK
	{ 0xff, 0xb6, 0x49 }, // ORANGE
	{ 0xff, 0x24, 0xb6 }, // LTPURPLE
	{ 0xff, 0xb6, 0x00 }, // DKORANGE
	{ 0xb6, 0x24, 0xff }, // DKPURPLE
	{ 0x00, 0xdb, 0xdb }, // DKCYAN
	{ 0xdb, 0xdb, 0x00 }, // DKYELLOW
	{ 0x95, 0x95, 0xff }, // BLUISH
	{ 0xff, 0x00, 0xff }, // PURPLE
};

enum {BLACK,WHITE,RED,GREEN,BLUE,CYAN,YELLOW,PINK,ORANGE,LTPURPLE,DKORANGE,
                DKPURPLE,DKCYAN,DKYELLOW,BLUISH,PURPLE};

#define UNUSED BLACK

#define COLOR_CODES 32

// NOTE:  This palette organization is different from Richard's.

static const BYTE _rgbColorTable[4 * COLOR_CODES] =
{
	/* charset A pallette A */  
	BLACK,BLACK,CYAN,CYAN,          /* Background, Unused, Letters, asterisks */   
	BLACK,YELLOW,RED,WHITE,         /* Background, Ship middle, Numbers/Ship, Ship edge */     
	BLACK,YELLOW,RED,WHITE,         /* Background, Ship middle, Ship, Ship edge/bullets */     
	BLACK,PINK,PURPLE,YELLOW,       /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,PINK,PURPLE,YELLOW,       /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,PINK,PURPLE,YELLOW,       /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,WHITE,PURPLE,YELLOW,      /* Background, Explosions */       
	BLACK,PURPLE,GREEN,WHITE,       /* Background, Barrier */
	/* charset B pallette A */
	BLACK,RED,BLUE,WHITE,           /* Background, Starfield */   
	BLACK,PURPLE,BLUISH,DKORANGE,   /* Background, Planets */     
	BLACK,DKPURPLE,GREEN,DKORANGE,  /* Background, Mothership: turrets, u-body, l-body */      
	BLACK,BLUISH,DKPURPLE,LTPURPLE, /* Background, Motheralien: face, body, feet */   
	BLACK,PURPLE,BLUISH,GREEN,      /* Background, Eagles: face, body, shell */   
	BLACK,PURPLE,BLUISH,GREEN,      /* Background, Eagles: face, body, feet */    
	BLACK,PURPLE,BLUISH,GREEN,      /* Background, Eagles: face, body, feet */    
	BLACK,PURPLE,BLUISH,GREEN,      /* Background, Eagles: face, body, feet */
	/* charset A pallette B */
	BLACK,BLUE,CYAN,CYAN,           /* Background, Unused, Letters, asterisks */   
	BLACK,YELLOW,RED,WHITE,         /* Background, Ship middle, Numbers/Ship, Ship edge */     
	BLACK,YELLOW,RED,WHITE,         /* Background, Ship middle, Ship, Ship edge/bullets */     
	BLACK,YELLOW,GREEN,PURPLE,      /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,YELLOW,GREEN,PURPLE,      /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,YELLOW,GREEN,PURPLE,      /* Background, Bird eyes, Bird middle, Bird Wings */       
	BLACK,WHITE,RED,PURPLE,         /* Background, Explosions */       
	BLACK,PURPLE,GREEN,WHITE,       /* Background, Barrier */
	/* charset B pallette B */
	BLACK,RED,BLUE,WHITE,           /* Background, Starfield */   
	BLACK,PURPLE,BLUISH,DKORANGE,   /* Background, Planets */     
	BLACK,DKPURPLE,GREEN,DKORANGE,  /* Background, Mothership: turrets, upper body, lower body */      
	BLACK,BLUISH,DKPURPLE,LTPURPLE, /* Background, Motheralien: face, body, feet */   
	BLACK,BLUISH,LTPURPLE,GREEN,    /* Background, Eagles: face, body, shell */   
	BLACK,BLUISH,LTPURPLE,GREEN,    /* Background, Eagles: face, body, feet */    
	BLACK,BLUISH,LTPURPLE,GREEN,    /* Background, Eagles: face, body, feet */    
	BLACK,BLUISH,LTPURPLE,GREEN,    /* Background, Eagles: face, body, feet */
};

static const SIZE _sizBuffer = { 26 * 8, 32 * 8 + 8};
static const RECT _rcDisplay = { 0, 8, 26 * 8, 256 - 8 };

static const DECODE_GRAPHICS dgsChars =
{
	0x00000001,
	{
		{ 8, 8 },
		256,
		8 * 8,
		2,
		{ 0, 8 * 8 * 256 },
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 7, 6, 5, 4, 3, 2, 1, 0 },
	},

	1,
	
	1, 
	
	{
		0,
	},
	0, 
	NULL
};

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);

static const EXECZ80_MEM_DESCRIPTOR rgmd[] = 
{
	{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
	{ 0, 0, 0, (EXECZ80_MEMREAD_PROC)0xFFFFFFFF, 0 },
};

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

HRESULT PhoenixInit(GAME_LOAD_INFO &gli)
{
	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(_rgpe, DIM(_rgpe), _rgbColorTable,
		COLOR_CODES * 4, gli.ppe, 256, gli.cpe);

	BYTE tmpstorage[0x2000];

	if ((DWORD)gli.pvContext == 0x00000000)
	{
		if (readroms(RAM, ROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;

		if (readroms(tmpstorage, GFXROMs, szPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;
	}
	else if ((DWORD)gli.pvContext == 0x00000001)
	{
		if (readroms(RAM, PleiadesROMs, szPleiadesPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;

		if (readroms(tmpstorage, PleiadesGFXROMs, szPleiadesPath, gli.pszMsg, gli.cchMsgMax) != 0)
			return E_FAIL;
	}

	//
	//  Decode characters
	//
	//  NOTE:  for some reason, can't use { { 8, 8 }, 512, ... } to decode charset B!!!
	//

	MyGfxDecode3(dgsChars, tmpstorage, _ms.pimgChars1);
	MyGfxDecode3(dgsChars, tmpstorage + 0x1000, _ms.pimgChars2);

	//
	//
	//
	
	EXECZ80_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pfnMemRead = M_RDMEM;
	cs.pfnMemWrite = M_WRMEM;
	cs.pmd = rgmd;
	
	HRESULT hr = EclCreateZ80(cs, _ms.pZ80);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

void PhoenixTerm()
{
	FreeImage(_ms.pimgChars1);
	FreeImage(_ms.pimgChars2);
	_ms.pZ80->Release();
}


/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	//
	//  are we reading dip switch memory?
	//
	
	if (A >= PHOENIX_IN0_PORT_START && A <= PHOENIX_IN0_PORT_END)
	{
		BYTE b = 0xFF;

		if (osd_key_pressed(OSD_KEY_LEFT))
			b ^= PHOENIX_IN0_P1_LEFT;

		if (osd_key_pressed(OSD_KEY_RIGHT))
			b ^= PHOENIX_IN0_P1_RIGHT;

		if (osd_key_pressed(OSD_KEY_3))
			b ^= PHOENIX_IN0_COIN1;

		if (osd_key_pressed(OSD_KEY_1))
			b ^= PHOENIX_IN0_1P_START;

		if (osd_key_pressed(OSD_KEY_2))
			b ^= PHOENIX_IN0_2P_START;

		if (osd_key_pressed(OSD_KEY_CONTROL))
			b ^= PHOENIX_IN0_P1_FIRE;

		if (osd_key_pressed(OSD_KEY_ALT))
			b ^= PHOENIX_IN0_P1_BARRIER;

		return b;
	}
	
	//
	//  are we reading dip switch memory?
	//
	
	if (A >= PHOENIX_DSW_PORT_START && A <= PHOENIX_DSW_PORT_END)
	{
		BYTE b = 0xEC;

		if (_ms.interrupt_set == 1)
			_ms.interrupt_set = 0;
		else
			b ^= 0x80;

		return b;
	}

	//
	//
	//

	return RAM[A];
}


/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	if (A < PHOENIX_END_OF_ROM)
		return;

	if (A >= SCROLL_REG_START && A < SCROLL_REG_START + SCROLL_REG_SIZE)
	{
		_ms.scrollreg = V;
	}
	else if (A >= VIDEO_REG_START && A < VIDEO_REG_START + VIDEO_REG_SIZE)
	{
//		(0xFF << 6 = 0xC0) >> 7 = 
		_ms.palreg = 16 * ((V << 6) >> 7); /* there must be a better way! */
		_ms.bankreg = (V << 7) >> 7;
	}

/*
	if (A == PHOENIX_SOUND_PORT_B) // A)
	{
		osd_sound((V & 0x0F) * 80 * 8);
	}
*/

	RAM[A] = V;
}

/*---------------------------------------------------------------------------

   6000     sound control A
         Bit   Used for
         ---------------
         0  ]
         1  ] frequency voice1
         2  ] 
         3  ]
         4   ] volume voice1 (probably)
         5   ] 
         6  ] melody module command:
         7  ] { <silence>, <for luise>, <???>, <alarm beep> }
   6000-63ff   sound control A (mirrored)
   
   6800     sound control B
         Bit   Used for
         ---------------
         0  ]
         1  ] frequency voice2
         2  ] 
         3  ]
         4   ] volume voice2 (probably)
         5   ] 
         6  ] noise channel (probably volume 0-3)
         7  ]
   6800-6bff   sound control B (mirrored)

---------------------------------------------------------------------------*/

void PhoenixUpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
	static const BYTE samples[][32] =
	{
		0x00,0x00,0x00,0x00,0x22,0x22,0x22,0x22,0x44,0x44,0x44,0x44,0x22,0x22,0x22,0x22,
		0x00,0x00,0x00,0x00,0xdd,0xdd,0xdd,0xdd,0xbb,0xbb,0xbb,0xbb,0xdd,0xdd,0xdd,0xdd
	};

	int freq = (RAM[PHOENIX_SOUND_PORT_B] & 0x0F) * 80 * 8;
	int iFreq = (freq / 80) & 0xFF;

	for (int i = 0; i < cbBuffer; i++)
	{
		char y = samples[0][(i * iFreq / 256) % 32] / 4;
			
		rgbBuffer[i] = y;
	}
}

#if 0

// from MAME...

/* I played around with these until they sounded ok */
static max_freq = 22500;
static dec_freq = 1500;

/* values needed by phoenix_sh_update */
static int sound_a_play = 0;
static int sound_a_adjust = -750;
static int sound_a_sine[4] = { 0, 250, 0, -250 };
static int sound_a_sine_adjust = 0;
static int sound_a_vol = 0;
static int sound_a_freq = 1000;

static int sound_b_play = 0;
static int sound_b_adjust = -750;
static int sound_b_vol = 0;
static int sound_b_freq = 1000;

static int noise_vol = 0;
static int noise_freq = 1000;

void phoenix_sound_control_a_w(int offset,int data)
{
        /* voice a */
        int freq = data & 0x0f;
        int vol = (data & 0x30) >> 4;

        /* noise */
        int noise = (data & 0xc0) >> 6;

        if (freq != sound_a_freq) sound_a_adjust = 0;
        sound_a_freq = freq;
        sound_a_vol = vol;

        if (freq != 0x0f)
        {
                osd_adjust_sample(0,max_freq-(dec_freq*freq),85*(3-vol));
                sound_a_play = 1;
        }
        else
        {
                osd_adjust_sample(0,1000,0);
                sound_a_play = 0;
        }

        if (noise_freq != 1500*(4-noise))
        {
                noise_freq = 1500*(4-noise);
                noise_vol = 85*noise;
        }

        if (noise) osd_adjust_sample(2,noise_freq,noise_vol);
        else
        {
                osd_adjust_sample(2,1000,0);
                noise_vol = 0;
        }
}

void phoenix_sound_control_b_w(int offset,int data)
{
        /* voice b */
        int freq = data & 0x0f;
        int vol = (data & 0x30) >> 4;

        /* melody - osd_play_midi anyone? */
        /* 0 - no tune, 1 - alarm beep?, 2 - even level tune, 3 - odd level tune */
        /*int tune = (data & 0xc0) >> 6;*/

        if (freq != sound_b_freq) sound_b_adjust = 0;
        sound_b_freq = freq;
        sound_b_vol = vol;

        if (freq != 0x0f)
        {
                osd_adjust_sample(1,max_freq-(dec_freq*freq),85*(3-vol));
                sound_b_play = 1;
        }
        else
        {
                osd_adjust_sample(1,1000,0);
                sound_b_play = 0;
        }
}



int phoenix_sh_start(void)
{
        osd_play_sample(0,Machine->drv->samples,32,1000,0,1);
        osd_play_sample(1,Machine->drv->samples,32,1000,0,1);
        osd_play_sample(2,&Machine->drv->samples[32],128,1000,0,1);
	return 0;
}



void phoenix_sh_update(void)
{
        sound_a_adjust++;
        sound_b_adjust++;
        sound_a_sine_adjust++;

        if (sound_a_adjust == 750) sound_a_adjust = -750;
        if (sound_b_adjust == 750) sound_b_adjust = -750;
        if (sound_a_sine_adjust == 4) sound_a_sine_adjust = 0;

        if (sound_a_play)
                osd_adjust_sample(0,max_freq-(dec_freq*sound_a_freq)-sound_a_adjust-sound_a_sine[sound_a_sine_adjust],85*(3-sound_a_vol));
        if (sound_b_play)
                osd_adjust_sample(1,max_freq-(dec_freq*sound_b_freq)-sound_b_adjust,85*(3-sound_b_vol));
        if (noise_vol)
        {
                osd_adjust_sample(2,noise_freq,noise_vol);
                noise_vol-=3;
        }
}


#endif

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int PhoenixExec()
{
	int cCycles = _ms.pZ80->Exec(12000);
	_ms.interrupt_set = 1;
	return cCycles;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void PhoenixRenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	LPBYTE pBase;
	int i;

	//
	//  Draw scrolling background layer of characters
	//

	pBase = &RAM[0x4800];

	int y = RAM[base + 0x1800] / 8;
	const int scan_line = RAM[base + 0x1800] % 8;

	for (int y1 = 0; y1 < 32; y1++)
	{
		for (int x = 0; x < 26; x++)
		{
			int sx = x * 8;
			int sy = 8 + (y1 * 8) - scan_line;

			const int offset_value = pBase[(25 - x) * 32 + y];
			const int color = ((256 + offset_value >> 5) | _ms.palreg) << 2;

			DrawChar(_ms.pimgChars2->rgbData, offset_value, color, sx, sy, rgbBuffer, lPitch);
		}
		
		y++;
		y = y % 32;
	}

	//
	//  Draw transparent top layer of characters
	//

	pBase = &RAM[base];

	for (i = 0; i < 32 * 26; i++)
	{
		const int offset_value = *pBase++;

		if (offset_value == 0)
			continue;

		const int color = ((offset_value >> 5) | _ms.palreg) << 2;

		int sx = 8 * (25 - (i / 32));
		int sy = 8 + 8 * (i % 32);

		DrawCharT(_ms.pimgChars1->rgbData, _ms.pimgChars1->rgbMask, offset_value, color, sx, sy, rgbBuffer, lPitch);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdPhoenix = 
{
	TEXT("Phoenix"),
	NULL,
	PhoenixInit,
	PhoenixExec,
	PhoenixRenderScene,
	PhoenixUpdateSound,
	PhoenixTerm
};

GAME_DESCRIPTOR _gmdPleiades = 
{
	TEXT("Pleiades"),
	(LPVOID)0x00000001,
	PhoenixInit,
	PhoenixExec,
	PhoenixRenderScene,
	PhoenixUpdateSound,
	PhoenixTerm
};

#if 0

// try something like this...

const BYTE *ptr = &RAM[base];

for (i = 0; i < 32 * 8; i += 8)
{
	for (j = 25 * 8 + 8; j >= 0 * 8 + 8; j -= 8)
	{
		const int c = *ptr++;

		// skip blanks in playfield
		if (c != 0x00)
		{
			const int color = ((offset_value >> 5) | _ms.palreg) << 2;

			DrawCharT(rgbBuffer, lPitch, j, i, c, color);
		}
	}
}

// or, trying doing that with a single loop variable ...

#endif
