/*---------------------------------------------------------------------------

	File:
		vanguard.cpp
	
	Abstract:
		Implements the "Vanguard" game driver functions.

	Revisions:
		15-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	All hardware, ROM locations, and graphics information was adopted from
	the Vanguard emulator by Brian Levine.

---------------------------------------------------------------------------*/

#include "pch.h"
#include <stdio.h>
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../inc/z80api.h"

extern EXEC_Z80_STRUCT _es; // hack until better 6502 interface is defined

namespace ns6502
{
#include "../m6502/old/6502.h"
}

extern "C"
{
void Reset6502(reg *Regs);
word M6502(reg Regs);
};

static BYTE _rgbChars[512][8][8];
static BYTE _rgbChars2[512][8][8];
static BYTE _rgbSprites[256][16][16];
static BYTE _rgbSpriteMask[256][16][16];

#include "blit.h"

inline static void DrawChar2(LPBYTE rgbBuffer, LONG lPitch, int X, int Y,
	int offset_value, int color)
{
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * Y + X];
	LPCDWORD pSrc = (LPCDWORD)&_rgbChars2[offset_value][0][0];

	// distribute color selector into all four bytes
	color |= (color << 8) | (color << 16) | (color << 24);

	lPitch = (lPitch / 4) - 1;

	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
}

static const TCHAR _szPath[] = "vanguard";

RAGE_ROMTABLE_BEGIN(_rgRoms)
	RAGE_ROMTABLE_ENTRY("sk4_ic07.bin", 0x4000, 0x1000) // code
	RAGE_ROMTABLE_ENTRY("sk4_ic08.bin", 0x5000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk4_ic09.bin", 0x6000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk4_ic10.bin", 0x7000, 0x1000) // code/sprites/tables

	RAGE_ROMTABLE_ENTRY("sk4_ic13.bin", 0x8000, 0x1000) // sprite graphics
	RAGE_ROMTABLE_ENTRY("sk4_ic14.bin", 0x9000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk4_ic15.bin", 0xa000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk4_ic16.bin", 0xb000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk5_ic50.bin", 0xc000, 0x0800) // character graphics
	RAGE_ROMTABLE_ENTRY("sk5_ic51.bin", 0xc800, 0x0800) // character graphics
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRoms)
	RAGE_ROMTABLE_ENTRY("sk4_ic13.bin", 0x0000, 0x1000)
	RAGE_ROMTABLE_ENTRY("sk5_ic50.bin", 0x1000, 0x0800)
	RAGE_ROMTABLE_ENTRY("sk5_ic51.bin", 0x1800, 0x0800)
RAGE_ROMTABLE_END

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

enum colors { Black=0, Blue, Green, Cyan, Red, Magenta, Brown, LGray,
              DGray, HBlue, HGreen, HCyan, HRed, HMagenta, Yellow, White };

static const PALETTEENTRY _rgpe[] =
{
	{ 0x00, 0x00, 0x00 },
	{ 0x00, 0x00, 0xC0 }, 
	{ 0x00, 0xC0, 0x00 },
	{ 0x00, 0xC0, 0xC0 },
	{ 0xC0, 0x00, 0x00 },
	{ 0xC0, 0x00, 0xC0 },
	{ 0xC0, 0xC0, 0x00 },
	{ 0x00, 0x00, 0x00 },

	{ 0x80, 0x80, 0x80 },
	{ 0x00, 0x00, 0xFF },
	{ 0x00, 0xFF, 0x00 },
	{ 0x00, 0xFF, 0xFF },
	{ 0xFF, 0x00, 0x00 },
	{ 0xFF, 0x00, 0xFF },
	{ 0xFF, 0xFF, 0x00 },
	{ 0xFF, 0xFF, 0xFF }
};

static const BYTE _rgbColorTable[] =
{
	// foreground
	
	Black, Cyan, Brown, White ,         //-- extra energy level
	Black, HMagenta, Blue, Red ,
	Black, HRed, Blue, Yellow ,         //-- special text, laser
	Black, Yellow, Blue, White ,
	Black, HMagenta, HGreen, HRed ,     //-- ship
	Black, Cyan, HBlue, White ,         //-- map, enemy
	Black, HMagenta, Red, Blue ,        //-- small text
	Black, HMagenta, Red, Blue,          //-- enemy under attack

	// background
	
	Black, HCyan, White, Blue ,          
	Black, White, HCyan, Red ,          
	Black, Cyan, Red, White ,           //-- energy
	Black, HCyan, White, HBlue ,        //-- logo, background
	Black, Red, Brown, Yellow ,         //-- enemy explosion
	Black, HBlue, HCyan, White ,
	Black, White, Yellow, Blue,        
	Black, Yellow, Blue, White,
};

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
	static const RECT _rcDisplay = { 0, 0, 256 - 16, 256 };

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(_rgpe, DIM(_rgpe), _rgbColorTable,
		DIM(_rgbColorTable), gli.ppe, 256, gli.cpe);

	if (readroms(RAM, _rgRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//  NOTE:  The ROMs are patched here to provide the interrupt vectors.
	//         There must be a ROM image that has these already in place!!!
	//
	
	RAM[0xFFFA] = 0x00; // nmi vector
	RAM[0xFFFB] = 0x74;

	RAM[0xFFFC] = 0x00; // reset vector
	RAM[0xFFFD] = 0x40;

	RAM[0xFFFE] = 0x00; // irq vector
	RAM[0xFFFF] = 0x70;

	BYTE tmpstorage[0x2000];

	if (readroms(tmpstorage, _rgGfxRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	static const MY_GFX_DECODE mgdChar =
	{
		{ 8, 8 },
		512,
		8 * 8,
		2,
		8 * 8 * 256,
		2,
		{ 7*8, 6*8, 5*8, 4*8, 3*8, 2*8, 1*8, 0*8 },
		{ 0, 1, 2, 3, 4, 5, 6, 7 }
	};

	// CONSIDER:  alloc _rgbChar/_rgbSprites dynamically based on MY_GFX_DECODE
	MyGfxDecode(tmpstorage + 0x1000, mgdChar, (LPBYTE)_rgbChars);

	MyGfxDecode(tmpstorage, mgdChar, (LPBYTE)_rgbChars2);

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	switch (A)
	{
		case 0x3104:
		case 0x3105:
		{
			BYTE b = 0x00;
			
			if (osd_key_pressed(OSD_KEY_S))
				b ^= 0x01;

			if (osd_key_pressed(OSD_KEY_W))
				b ^= 0x02;
 
			if (osd_key_pressed(OSD_KEY_D))
				b ^= 0x04;

			if (osd_key_pressed(OSD_KEY_A))
				b ^= 0x08;
	  
			if (osd_key_pressed(OSD_KEY_DOWN))
				b ^= 0x10;
	  
			if (osd_key_pressed(OSD_KEY_UP))
				b ^= 0x20;
	  
			if (osd_key_pressed(OSD_KEY_RIGHT))
				b ^= 0x40;
	  
			if (osd_key_pressed(OSD_KEY_LEFT))
				b ^= 0x80;
	  
			return b;
		}

		case 0x3107:
		{
			BYTE b = 0x00;
			
			if (osd_key_pressed(OSD_KEY_3))
				b ^= 0x01;
	  
			if (osd_key_pressed(OSD_KEY_4))
				b ^= 0x02;

			if (osd_key_pressed(OSD_KEY_2))
				b ^= 0x40;

			if (osd_key_pressed(OSD_KEY_1))
				b ^= 0x80;

			return b;
		}

		default:
			return(RAM[A]);
	}
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	// don't allow code to overwrite itself
	if (A >= 0x4000)
		return;

	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int Interrupt(void)
{
	if (!CheckMessages())
		CPURunning = 0;

	Throttle(60);

	if (osd_key_pressed(OSD_KEY_3))
	{
		RAM[0x3107] |= 0x01;
		
		return INT_NMI;
	}
	else if (osd_key_pressed(OSD_KEY_4))
	{
		RAM[0x3107] |= 0x02;
		
		return INT_NMI;
	}   

	return INT_IRQ;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

// TODO:  call higher-level Exec function that handles the common parts
//        of Interrupt, pass it our CPU speed and frame rate

static HRESULT Run(LPVOID pvContext)
{
	_es.pfnReadMemory  = M_RDMEM;
	_es.pfnWriteMemory = M_WRMEM;
	_es.pfnInterrupt   = Interrupt;

	reg Regs;

	Reset6502(&Regs);

	M6502(Regs);

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static HRESULT RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i;
	
/*
	if (GetAsyncKeyState('C') < 0)
	{
		if (GetAsyncKeyState('1') < 0)
		{
			for (i = 0; i < DIM(_rgbChars); i++)
			{
				DrawChar(rgbBuffer, lPitch, 8 * (i % 32), 8 * (i / 32), i, 0);
			}
		}

		if (GetAsyncKeyState('2') < 0)
		{
			for (i = 0; i < DIM(_rgbChars2); i++)
			{
				DrawChar2(rgbBuffer, lPitch, 8 * (i % 32), 8 * (i / 32), i, 0);
			}
		}

		return S_OK;
	}
*/

	//
	//  Draw scrolling opaque background
	//
	
	int sx = RAM[0x3300];
	int sy = 255 - RAM[0x3200];
	
	for (i = 0; i < 32 * 32; i++)
	{
		DrawChar((LPBYTE)_rgbChars, RAM[0x0800 + i], (RAM[0x0C00 + i] & 0x70) >> 1,
			(sx + (8 * (31 - (i / 32)))) % 256, (sy + (8 * (i % 32))) % 256,
			rgbBuffer, lPitch);
	}

	//
	//  Draw overlayed character graphics
	//

	for (i = 0; i < 32 * 32; i++)
	{
		int c = RAM[0x0400 + i];
		
		if (c == 48)
			continue;

		DrawChar2(rgbBuffer, lPitch, 8 * (31 - (i / 32)), 8 * (i % 32),
			c, RAM[0x0C00 + i] & 0x07);
	}

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

GAME_DESCRIPTOR _gmdVanguard = 
{
	TEXT("Vanguard"),
	Init,
	Run,
	RenderScene,
	NULL,
	NULL,
	NULL,
	NULL
};

