/*---------------------------------------------------------------------------

	File:
		venture.cpp
	
	Abstract:
		Implements the "Venture" game driver functions.

	Revisions:
		25-Apr-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Hardware information adopted from the Venture driver for MAME.

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "game.h"
#include "game1.h"
#include "../m6502/exec6502.h"

static IExec6502 *_p6502;

extern BYTE _rgbRAM[0x10000];

static BYTE _rgbChars[256][8][8];
static BYTE _rgbSprites[512][16][16];
static BYTE _rgbSpriteMask[512][16][16];

static const TCHAR _szPath[] = "venture";

RAGE_ROMTABLE_BEGIN(_rgRoms)
	RAGE_ROMTABLE_ENTRY("13A-CPU", 0x8000, 0x1000)
	RAGE_ROMTABLE_ENTRY("12A-CPU", 0x9000, 0x1000)
	RAGE_ROMTABLE_ENTRY("11A-CPU", 0xA000, 0x1000)
	RAGE_ROMTABLE_ENTRY("10A-CPU", 0xB000, 0x1000)
	RAGE_ROMTABLE_ENTRY("9A-CPU",  0xC000, 0x1000)
	RAGE_ROMTABLE_ENTRY("8A-CPU",  0xD000, 0x1000)
	RAGE_ROMTABLE_ENTRY("7A-CPU",  0xE000, 0x1000)
	RAGE_ROMTABLE_ENTRY("6A-CPU",  0xF000, 0x1000)
RAGE_ROMTABLE_END

RAGE_ROMTABLE_BEGIN(_rgGfxRoms)
	RAGE_ROMTABLE_ENTRY("11D-CPU", 0x0000, 0x0800)
RAGE_ROMTABLE_END

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static const PALETTEENTRY palette[] =
{
	{ 0x00, 0x00, 0x00 }, // black      
	{ 0x94, 0x00, 0xd8 }, // darkpurple 
	{ 0xd8, 0x00, 0x00 }, // darkred    
	{ 0xf8, 0x64, 0xd8 }, // pink       
	{ 0x00, 0xd8, 0x00 }, // darkgreen  
	{ 0x00, 0xf8, 0xd8 }, // darkcyan   
	{ 0xd8, 0xd8, 0x94 }, // darkyellow 
	{ 0xd8, 0xf8, 0xd8 }, // darkwhite  
	{ 0xf8, 0x94, 0x44 }, // orange     
	{ 0x00, 0x00, 0xd8 }, // blue   
	{ 0xf8, 0x00, 0x00 }, // red    
	{ 0xff, 0x00, 0xff }, // purple 
	{ 0x00, 0xf8, 0x00 }, // green  
	{ 0x00, 0xff, 0xff }, // cyan   
	{ 0xf8, 0xf8, 0x00 }, // yellow 
	{ 0xff, 0xff, 0xff }, // white
};

enum
{
	black, darkpurple, darkred, pink, darkgreen, darkcyan, darkyellow,
		darkwhite, orange, blue, red, purple, green, cyan, yellow, white
};

static const unsigned char venture_colortable[] =
{
	black, yellow,
	black, cyan,
	black, white,
	black, purple,
	black, white,

	black, cyan,
	black, yellow,
};

static BYTE M_RDMEM(WORD A);
static void M_WRMEM(WORD A, BYTE);

static HRESULT Init(GAME_LOAD_INFO &gli)
{
	static const SIZE _sizBuffer = { 256 + 16, 256 + 16};
	static const RECT _rcDisplay = { 0, 0, 256 - 16, 256 };

	gli.sizBuffer = _sizBuffer;
	gli.rcDisplay = _rcDisplay;

	BuildCompositePalette(palette, DIM(palette), venture_colortable,
		DIM(venture_colortable), gli.ppe, 256, gli.cpe);

	if (readroms(RAM, _rgRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	BYTE tmpstorage[0x1000];

	if (readroms(tmpstorage, _rgGfxRoms, _szPath, gli.pszMsg, gli.cchMsgMax) != 0)
		return E_FAIL;

	//
	//
	//
	
	static const MY_GFX_DECODE mgdChar =
	{
		{ 8, 8 },
		256,
		8 * 8,
		1,
		0,
		1,
		{ 0, 1, 2, 3, 4, 5, 6, 7 },
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8 },
	};

	// CONSIDER:  alloc _rgbChar/_rgbSprites dynamically based on MY_GFX_DECODE
//	MyGfxDecode(tmpstorage, mgdChar, (LPBYTE)_rgbChars);
	MyGfxDecode2(tmpstorage, mgdChar, 1, NULL, (LPBYTE)_rgbChars, NULL);

	//
	//
	//
	
	static const MY_GFX_DECODE mgdSprites =
	{
		{ 16, 16 },
		64,
		32 * 8,
		1,
		0,
		1,
		{ 0, 1, 2, 3, 4, 5, 6, 7, 16*8+0, 16*8+1, 16*8+2, 16*8+3, 16*8+4, 16*8+5, 16*8+6, 16*8+7},
		{ 0*8, 1*8, 2*8, 3*8, 4*8, 5*8, 6*8, 7*8, 8*8, 9*8, 10*8, 11*8, 12*8, 13*8, 14*8, 15*8},
	};

	static const int rgiMirror[] = {
		0 * mgdSprites.cImage,
		1 * mgdSprites.cImage,
		2 * mgdSprites.cImage,
		3 * mgdSprites.cImage
	};

	static const int cImageInc = 1;

	MyGfxDecode2(tmpstorage, mgdSprites, cImageInc, rgiMirror,
		(LPBYTE)_rgbSprites, (LPBYTE)_rgbSpriteMask);

	//
	//
	//

	static const EXEC6502_MEM_DESCRIPTOR rgmd[] = 
	{
		{ 0x0000, 0xFFFF, 0, M_RDMEM, M_WRMEM },
		{ 0, 0, 0, (EXEC6502_MEMREAD_PROC)0xFFFFFFFF, 0 },
	};

	EXEC6502_CREATE cs;

	ZERO_INIT(cs);

	cs.rgbMemory = RAM;
	cs.pmd = rgmd;

	HRESULT hr;
	
	hr = EclCreate6502(cs, _p6502);

	if (FAILED(hr))
		return hr;

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static BYTE M_RDMEM(WORD A)
{
	if (A == 0x5100)
	{
		BYTE b = 0xB0;

		if (osd_key_pressed(OSD_KEY_4))
			b ^= 0x01;

		return b;
	}

	if (A == 0x5101)
	{
		BYTE b = 0xFF;

		if (osd_key_pressed(OSD_KEY_1))
			b ^= 0x01;

		if (osd_key_pressed(OSD_KEY_2))
			b ^= 0x02;

		if (osd_key_pressed(OSD_KEY_RIGHT))
			b ^= 0x04;

		if (osd_key_pressed(OSD_KEY_LEFT))
			b ^= 0x08;

		if (osd_key_pressed(OSD_KEY_CONTROL))
			b ^= 0x10;

		if (osd_key_pressed(OSD_KEY_UP))
			b ^= 0x20;

		if (osd_key_pressed(OSD_KEY_DOWN))
			b ^= 0x40;

		if (osd_key_pressed(OSD_KEY_3))
			b ^= 0x80;

		return b;
	}

	if (A == 0x5103)
	{
		BYTE b = 0x00;

		if (osd_key_pressed(OSD_KEY_7))
			b ^= 0x04;

		if (osd_key_pressed(OSD_KEY_7))
			b ^= 0x10;

		if (osd_key_pressed(OSD_KEY_7))
			b ^= 0x20;

		if (osd_key_pressed(OSD_KEY_7))
			b ^= 0x40;

		return b;
	}

	if (A == 0x5213)
	{
		BYTE b = 0xFF;

		if (osd_key_pressed(OSD_KEY_Z))
			b ^= 0x01;

		if (osd_key_pressed(OSD_KEY_X))
			b ^= 0x02;

		if (osd_key_pressed(OSD_KEY_E))
			b ^= 0x04;

		if (osd_key_pressed(OSD_KEY_C))
			b ^= 0x08;

		return b;
	}

	return(RAM[A]);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void M_WRMEM(WORD A, BYTE V)
{
	RAM[A] = V;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static int Exec()
{
	int cCycles;
	
	cCycles = _p6502->Exec(16667); // 1 MHz, 60 fps...  16667 cycles/frame
	cCycles += _p6502->Irq(); 

	return cCycles;
}

static void Term()
{
	_p6502->Release();
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void UpdateSound(LPBYTE rgbBuffer, LONG cbBuffer)
{
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void RenderScene(LPBYTE rgbBuffer, LONG lPitch)
{
	int i, j;

/*
	for (i = 0; i < 256; i++)
	{
		DrawChar((LPBYTE)_rgbChars, i, 0, 8 * (i % 32), 8 * (i / 32), rgbBuffer, lPitch);
	}

	return;
*/
	
	const BYTE *pBase = &RAM[0x4000];

	for (i = 0; i < 32 * 32; i++)
	{
		DrawChar((LPBYTE)_rgbChars, pBase[i], 0, // ((pBase[i] >> 6) + 1) << 1,
			8 * (i % 32), 8 * (i / 32), rgbBuffer, lPitch);
	}

/*
	for (i = 0; i < 16; i++)
	{
		int ypos = RAM[SPRITE_YPOS + i];

		if (ypos >= 0xF8)
			continue;

		ypos = 247 - ypos;

		const int xpos = 238 - RAM[SPRITE_XPOS + i];
		const int offset = RAM[SPRITE_OFFSET + i];
		const int color = 0; // RAM[SPRITE_COL + i] & 0x0F;

		DrawSprite(rgbBuffer, lPitch, xpos, ypos, offset, color);
	}
*/
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

#define DECLARE_GAME_DESCRIPTOR(name,title,ctx) \
	static HRESULT Init(GAME_LOAD_INFO &gli); \
	/*const*/ GAME_DESCRIPTOR (name) = { (title), (ctx), Init, Exec, RenderScene, UpdateSound, Term };

DECLARE_GAME_DESCRIPTOR(_gmdVenture, TEXT("Venture"),LPVOID(0));

