/*---------------------------------------------------------------------------
	
	File:
		Misc.h

	Abstract:
		Miscellaneous utility macros and templates.

	Revisions:
		07-Nov-95 EAM Created

---------------------------------------------------------------------------*/

#ifndef _MISC_H_
#define _MISC_H_

#include <stdio.h>

//
//  Validate arguments of OLE functions (or any function returning HRESULT)
//

#define VALIDATE_ARGS(e) \
	if (!(e)) { \
		_RPT1(_CRT_ERROR, "invalid arg(s): should be '%s'\n", #e); \
		return E_INVALIDARG; }

#ifdef _DEBUG
	#define DBG_CHECK_FLAGS(v,r) if ((v) & (r)) _RPT1(_CRT_WARN, "Invalid flags %08X\n", v)
#else
	#define DBG_CHECK_FLAGS(v,r) (void)0
#endif

//
//  Not-implemented macro for OLE functions (or any function returning HRESULT)
//

#define NOT_IMPLEMENTED(fn) \
	_RPT1(_CRT_ERROR, "feature not implemented: '%s'\n", fn); \
	return E_NOTIMPL;

//
//  Array dimension
//

#define DIM(rg) (sizeof(rg) / sizeof(rg[0]))

template <class T> int Dim(const T &rg)
{
	return sizeof(rg) / sizeof(rg[0]);
}

//
//  Value bounds
//

#define BOUND(val,lo,hi) ((val) >= (lo) && (val) <= (lo))

template <class T> BOOL BoundInclusive(const T &val, const T &lo, const T &hi)
{
	return (val >= lo && val <= hi);
}

template <class T> BOOL BoundExclusive(const T &val, const T &lo, const T &hi)
{
	return (val > lo && val < hi);
}

template <class T> void Copy(const T& tSrc, T& tDest)
{
	CopyMemory(&tDest, &tSrc, sizeof(tSrc));
}
	
#define ILLEGAL_CASE _ASSERT(FALSE); break

// disable "zero-sized array" warning
#define DISABLE_ARRAY_WARNING warning(disable:4200)
#define RESTORE_ARRAY_WARNING warning(default:4200)

#ifdef _DEBUG
#define VERIFY(e) _ASSERT(e)
#else
#define VERIFY(e) e
#endif

inline void ConvRect(const RECT &rcSrc, int cxMul, int cyMul,
	int cxDiv, int cyDiv, RECT &rcDest)
{
	rcDest.left   = MulDiv(rcSrc.left,   cxMul, cxDiv);
	rcDest.top    = MulDiv(rcSrc.top,    cyMul, cyDiv);
	rcDest.right  = MulDiv(rcSrc.right,  cxMul, cxDiv);
	rcDest.bottom = MulDiv(rcSrc.bottom, cyMul, cyDiv);
}

//
//  ex: FIELD_SIZEOF(POINT, x) is sizeof(int)
//

#define FIELD_SIZEOF(type, field)    (sizeof(((type*)0)->field))

#define ZERO_INIT(data) ZeroMemory(&data, sizeof(data))

//
//  TStr - template for fixed-length text strings.  Completely superseeds
//  local buffers (ex: TCHAR szBuf[32]) with no loss in performance.
//

template <UINT cch> class TStr
{
public:
	TStr()
	{
		m_rgch[0] = TEXT('\0');
		m_cch = 0;
	}

	TStr(LPCTSTR pszFmt, ...)
	{
		va_list marker;
		
		va_start(marker, pszFmt);

		int cchOut = _vsnprintf(m_rgch, cch, pszFmt, marker);

		if (cchOut >= 0)
			m_cch = cchOut;
	}

	void Format(LPCTSTR pszFmt, ...)
	{
		va_list marker;
		
		va_start(marker, pszFmt);

		int cchOut = _vsnprintf(m_rgch, cch, pszFmt, marker);

		if (cchOut >= 0)
			m_cch = cchOut;
	}

	void operator=(LPCTSTR psz)
	{
		Format("%s", psz);
	}

	void operator+=(LPCTSTR psz)
	{
		int cchOut = _snprintf(m_rgch + m_cch, cch - m_cch, "%s", psz);

		if (cchOut >= 0)
			m_cch += cchOut;
	}

	operator LPTSTR()
	{
		return &m_rgch[0];
	}

	operator UINT() const
	{
		return m_cch;
	}

	BOOL operator ==(LPCTSTR psz) const
	{
		return (lstrcmpi(psz, m_rgch) == 0);
	}

private:
	UINT m_cch;
	TCHAR m_rgch[cch];
};

//
//
//

class CWaitCursor
{
public:
	CWaitCursor()
	{
		m_hPrevCursor = SetCursor(LoadCursor(NULL, IDC_WAIT));
	}

	~CWaitCursor()
	{
		SetCursor(m_hPrevCursor);
	}

private:
	HCURSOR m_hPrevCursor;
};

#define WAIT_CURSOR CWaitCursor dummy_wait_cursor

//
//
//

#endif // _MISC_H_

