/*---------------------------------------------------------------------------

	File:
	
	Abstract:

	Revisions:
		25-Mar-97 EAM Created

	Copyright (C) 1997 Edward Massey

	- graphics decoding
	- palette composition and color PROM decoding
	- building interleaved sprite data
	- pre-rotating/mirroring sprites

---------------------------------------------------------------------------*/

#include "pch.h"
#include "../inc/graphics.h"
#include "misc.h"
#include "../inc/game1.h"

/*---------------------------------------------------------------------------

	GenSprites
	
	Generate sprite images and masks from character images.  Generates four
	images, one for each direction.

	Since the games in this family use similar graphics hardware, they all
	share this routine.  Each sprite has the following attributes:

	- horizontal and vertical screen position (in pixels)
	- mirror horizontal and mirror vertical flags
	- image index (6 bits -> 64 images)
	- color scheme (16 available)
	- bits (each pixels is two bits)

	Given this information, it is possible to do some major optimizations.
	Since the mirror bits are the upper two bits of the byte that also 
	contains the image index, we can simply use the entire byte directly as
	an image index where each image is mirrored in four positions.

	When both pixel bits are zero, that pixel is transparent.  To prevent
	doing a comparison inside the innermost sprite drawing loop, we generate
	a mask.

---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------

	MyGfxDecode3

	TODO:  add support for Phoenix (makes two decode calls for one set).
	TODO:  just make this routine IMAGE *MyGfxDecode3(DECODE_GRAPHICS &dgs)
	TODO:  add a Load and Decode routine that takes a ROM load list
	TODO:  implement the interleaved (mask/data) decode and blit
	TODO:  allow centipede graphics (only two mirrored images)
	TODO:  allow decode of char's at runtime (following write to programmable
		character memory...  venture, carnival)

parameters...

number of sprites
generate [interleaved] mask data
layout of sprite indexes (ex: iImage + 64 * iDir or 4 * iImage + iDir)
    layout of mirroring (should we do any mirroring)
ROMs (filenames, locations)
number of bits to left shift the sprite data in the destination buffer
	(may help with palette optimizations later)
color table info (may allow us to do some palette optimizations)

---------------------------------------------------------------------------*/

static void DecodePixel(const DECODE_GRAPHICS &dgs, const BYTE *rgbSource,
	int iImage, int i, int j, IMAGE *pImage)
{
	int iDir, iPlane;

	const int cbImage = dgs.mgd.sizImage.cx * dgs.mgd.sizImage.cy;
	BOOL fTransparent = (dgs.dwFlags & DGF_TRANSPARENT) != 0;

	//
	//
	//
	
	BYTE b = 0;

	for (iPlane = 0; iPlane < dgs.mgd.cPlane; iPlane++)
	{
		int offs = iImage * dgs.mgd.cbImagePitch + 
			dgs.mgd.rgcbPlanePitch[iPlane];
		
		int bitnum = offs + dgs.mgd.yoffset[i] + dgs.mgd.xoffset[j];
		
		b = (b << 1);

		int bits = rgbSource[bitnum / 8];

		bitnum %= 8;

		while (bitnum-- > 0)
			bits <<= 1;

		if (bits & 0x80)
			b |= 1;
	}

	//
	//
	//

	int sx, sy;

	// TODO: int sx[4], sy[4];
	
	for (iDir = 0; iDir < dgs.cMirror; iDir++)
	{
		switch (iDir)
		{
			case 0:
				sx = j;
				sy = i;
				break;

			case 1:
				sx = j;
				sy = (dgs.mgd.sizImage.cy - 1) - i;
				break;

			case 2:
				sx = (dgs.mgd.sizImage.cx - 1) - j;
				sy = i;
				break;

			case 3:
				sx = (dgs.mgd.sizImage.cx - 1) - j;
				sy = (dgs.mgd.sizImage.cy - 1) - i;
				break;

			default:
				_ASSERT(FALSE);
				break;
		}

		int iData = ((dgs.cImageInc * iImage + dgs.rgiMirror[iDir]) * cbImage)
			+ dgs.mgd.sizImage.cy * sy + sx;

		int iMask = iData;

		pImage->rgbData[iData] = b << dgs.cBitShift;

		if (fTransparent != NULL)
		{
			if (dgs.pnTransparent != NULL)
				b = dgs.pnTransparent[b];

			pImage->rgbMask[iMask] = b ? 0x00 : 0xFF;
		}
	}
}

HRESULT MyGfxDecode3(const DECODE_GRAPHICS &dgs, const BYTE *rgbSource, IMAGE *&pImage)
{
	_ASSERT(dgs.cMirror >= 1 && dgs.cMirror <= 4);
	_ASSERT(rgbSource != NULL);
	_ASSERT(dgs.cImageInc > 0);

	const int cbImage = dgs.mgd.sizImage.cx * dgs.mgd.sizImage.cy;
	const int cbImages = dgs.mgd.cImage * dgs.cMirror * cbImage;

	BOOL fTransparent = (dgs.dwFlags & DGF_TRANSPARENT) != 0;

	pImage = (IMAGE*)calloc(1, sizeof(IMAGE) + cbImages * (fTransparent ? 2 : 1));

	if (pImage == NULL)
		return E_OUTOFMEMORY;

	pImage->cImages = dgs.mgd.cImage * dgs.cMirror;
	pImage->sizImage = dgs.mgd.sizImage;
	pImage->dwFlags = dgs.dwFlags;
	pImage->rgbData = (LPBYTE)(pImage + 1);
	pImage->rgbMask = fTransparent ? pImage->rgbData + cbImages : NULL;

	//
	//  Decode each image...
	//
	
	int iImage, i, j;

	for (iImage = 0; iImage < dgs.mgd.cImage; iImage++)
	{
		for (i = 0; i < dgs.mgd.sizImage.cy; i++)
		{
			for (j = 0; j < dgs.mgd.sizImage.cx; j++)
			{
				DecodePixel(dgs, rgbSource, iImage, i, j, pImage);
			}
		}
	}

	return S_OK;
}

HRESULT MyGfxDecode4(const DECODE_GRAPHICS2 &dgs, IMAGE *&pImage)
{
	BYTE tmpstorage[0x10000]; // ATTN:  hope this is big enough

	LPBYTE dest = tmpstorage;

	const RomModule * romp  = NULL; // dgs.rgRomModule;

	while (romp->name)
	{
		FILE *f = fopen(romp->name, "rb");
		
		if (f != NULL)
		{
			if (fread(dest + romp->offset, 1, romp->size, f) != romp->size)
			{
			}

			fclose(f);
		}
		
		romp++;
	}

	return MyGfxDecode3(dgs, tmpstorage, pImage);
}

HRESULT FreeImage(IMAGE *pImage)
{
	if (pImage != NULL)
		free(pImage);

	return S_OK;
}

/*

store the following information somewhere else besides in the code

registry

ROMs: filename, seek, load address, size
DIP switch settings
hiscore table
control/memory map
size of screen buffer/visible rectangle
frame rate
interrupt rate

resources

palettes
sound envelopes

---------------------------------------------------

the graphics routines need to handle the following types of chars/sprites:

8x8x2 opaque characters, color depends on selectable scheme (galaxian)
8x8x2 opaque characters, color depends on character itself (dkong)
8x8x2 transparent characters, color depends character and palette (phoenix)
8x8x2 opaque chars, rotated (time pilot, gyruss)
16x16x2 ()
16x16x3	(mbros)
32x32x3	(zaxxon)
16x8x2 centipede
16x8x4 gyruss

*/

void BuildCompositePalette(const PALETTEENTRY *ppe, int cpe,
	const BYTE *pclt, UINT cclt, PALETTEENTRY *ppeOut, int cpeOutMax,
	UINT &cpeOut)
{
	for (cpeOut = 0; cpeOut < cclt; cpeOut++)
		ppeOut[cpeOut] = ppe[pclt[cpeOut]];
}

#ifdef _DEBUG

void DrawChar(LPCBYTE rgbSource, int offset_value,
	int color, int X, int Y, LPBYTE rgbBuffer, LONG lPitch)
{
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * Y + X];
	LPCDWORD pSrc = (LPCDWORD)&rgbSource[offset_value * 8 * 8];

	// distribute color selector into all four bytes

	color |= (color << 8) | (color << 16) | (color << 24);

	lPitch = (lPitch / 4) - 1;

	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc++); pDest += lPitch;
	*pDest++ = (color | *pSrc++); *pDest = (color | *pSrc);
}

/*

Draw an opaque 8x8 image using only the color information contained in the
image data itself.

This routine is used by Donkey Kong, Donkey Kong Jr., and Mario Bros., since
the character backgrounds color is defined by the character itself, instead
of a color VRAM bank, as is typical in the other games.

*/

void DrawChar(LPCBYTE rgbSource, int offset_value,
	int X, int Y, LPBYTE rgbBuffer, LONG lPitch)
{
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * Y + X];
	LPCDWORD pSrc = (LPCDWORD)&rgbSource[offset_value * 8 * 8];

	lPitch = (lPitch / 4) - 1;

	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc++; pDest += lPitch;
	*pDest++ = *pSrc++; *pDest = *pSrc;
}

// consider: use a function template, parameterized with number of color
// schemes, bit depth of sprites, put in header file (inline)

void DrawSprite(LPCBYTE rgbData, LPCBYTE rgbMask,
	int n, int color, int x, int y, LPBYTE rgbBuffer, LONG lPitch)
{
	//
	//
	//
	
	color |= (color << 8) | (color << 16) | (color << 24);

	//
	//
	//
	
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&rgbData[n * 16 * 16];
	LPCDWORD pMask = (LPCDWORD)&rgbMask[n * 16 * 16];

	DWORD dwTemp;
	DWORD dwMask;
	
	//
	//  Draw sprite, with transparency, and color scheme adjusted
	//  to offscreen buffer, 32-bits at a time
	//
	//  CONSIDER:  unroll the loop
	//
	
	lPitch = (lPitch / 4) - 4;
	
#if 0
	for (int i = 0; i < 16; i++)
	{
		dwTemp = *pDest;
		dwMask = *pSrc++;
		dwTemp &= dwMask;
		dwMask = ~dwMask;
		dwTemp |= (color | *pSrc++) & ~dwMask);
		*pDest++ = dwTemp;
	}
#else
	for (int i = 0; i < 16; i++)
	{
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;
	}
#endif
}

void DrawSprite2(LPBYTE rgbBuffer, LONG lPitch, int x, int y,
	int n, int color, LPCBYTE rgbData, LPCBYTE rgbMask)
{
	//
	//
	//
	
	color |= (color << 8) | (color << 16) | (color << 24);

	//
	//
	//
	
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&rgbData[n * 32 * 32];
	LPCDWORD pMask = (LPCDWORD)&rgbMask[n * 32 * 32];
	
	//
	//  Draw sprite, with transparency, and color scheme adjusted
	//  to offscreen buffer, 32-bits at a time
	//
	//  CONSIDER:  unroll the loop
	//
	
	lPitch = (lPitch / 4) - 8;
	
	for (int i = 0; i < 8; i++)
	{
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		pDest += lPitch;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;

		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
		pDest++; pMask++; pSrc++;
		*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	}
}

void DrawCharT(LPCBYTE rgbData, LPCBYTE rgbMask,
	int n, int color, int x, int y, LPBYTE rgbBuffer, LONG lPitch)
{
	LPDWORD pDest = (LPDWORD)&rgbBuffer[lPitch * y + x];
	LPCDWORD pSrc = (LPCDWORD)&rgbData[n * 8 * 8];
	LPCDWORD pMask = (LPCDWORD)&rgbMask[n * 8 * 8];

	color |= (color << 8) | (color << 16) | (color << 24);

	lPitch = (lPitch / 4) - 1;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;

	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest++; pMask++; pSrc++;
	*pDest = (*pDest & *pMask) | ((color | *pSrc) & ~(*pMask));
	pDest += lPitch; pMask++; pSrc++;
}

#endif

