/*---------------------------------------------------------------------------

	File:
		script.cpp
	
	Abstract:
		Horrible recursive descent parser to process game script for MageX.

	Revisions:

	Copyright (C) 1997 Edward Massey

---------------------------------------------------------------------------*/

#include "pch.h"
#include "misc.h"
#include "script.h"
#include "debug.h"

struct PARSER_TABLE
{
	int nState;
	int nInput;
	int nAction;
	int nStateNext;
};

static int _lineno = 1;

static const PARSER_TABLE parser_table[] =
{
	{ 0, EOF, 0, 0 },
	{ 0, ' ', 0, 0 },
	{ 0, '\t', 0, 0 },
	{ 0, '\r', 0, 0 },
	{ 0, '\n', 0, 0 },
};

enum
{
	TT_INVALID,
	TT_GAME,
	TT_IDENTIFIER,
	TT_LBRACE,
	TT_RBRACE,
	TT_TITLE,
	TT_STRING,
	TT_PATH,
	TT_DRIVER,
	TT_ROMSET,
	TT_ROM,
	TT_INTEGER,
	TT_LPAREN,
	TT_RPAREN,
	TT_SEMICOLON,
	TT_COLON,
	TT_COMMA,
	TT_NONE
};

static const LPCTSTR _pszTokenTypeNames[] = 
{
	"(invalid)",
	"game",
	"identifier",
	"{",
	"}",
	"title",
	"string",
	"path",
	"driver",
	"romset",
	"rom",
	"integer",
	"lparen",
	"rparen",
	";",
	":",
	","
};

struct KEYWORD_DEF
{
	TCHAR szKeyword[16];
	int   nTokenType;
};

struct TOKEN
{
	int nType;

//	union
//	{
		TCHAR szValue[128];
		int   nValue;
//	};
};

static const KEYWORD_DEF _rgKeywordDefs[] = 
{
	{ "game",   TT_GAME    },
	{ "title",  TT_TITLE   },
	{ "path",   TT_PATH    },
	{ "driver", TT_DRIVER  },
	{ "romset", TT_ROMSET  },
	{ "rom",    TT_ROM     },
	{ "",       TT_INVALID },
};

int GetKeywordTokenType(LPCTSTR pszKeyword)
{
	const KEYWORD_DEF *pkd = _rgKeywordDefs;

	while (pkd->nTokenType != TT_INVALID)
	{
		if (lstrcmpi(pszKeyword, pkd->szKeyword) == 0)
			break;

		pkd++;
	}

	return pkd->nTokenType;
}

int GetToken(FILE *pf, TOKEN &token)
{
	int nState = 0;

	LPTSTR pszValue = token.szValue;

	token.nType = TT_INVALID;
	*pszValue = 0;
	token.nValue = 0;
	
	for (;;)
	{
		const int ch = fgetc(pf);

		if (ch == '\n')
			_lineno++;

		if (ch == EOF)
			return TT_NONE;

		switch (nState)
		{
			case 0:
			{
				if (isspace(ch))
					break; 

				if (__iscsymf(ch))
				{
					*pszValue++ = ch;
					*pszValue = 0;
					nState = 1;
					break;
				}

				if (ch == '{')
					return TT_LBRACE;

				if (ch == '}')
					return TT_RBRACE;

				if (ch == '\"')
				{
					nState = 2;
					*pszValue = 0;
					break;
				}

				if (ch == '(')
					return TT_LPAREN;

				if (ch == ')')
					return TT_RPAREN;

				if (ch == ';')
					return TT_SEMICOLON;

				if (ch == ':')
					return TT_COLON;

				if (ch == ',')
					return TT_COMMA;

				if (isdigit(ch))
				{
					token.nValue = ch - '0';
					nState = 3;
					break;
				}

				if (ch == '/')
				{
					nState = 5;
					break;
				}

				return TT_INVALID;

				break;
			}

			case 1:
			{
				if (__iscsym(ch))
				{
					*pszValue++ = ch;
					*pszValue = 0;
					break;
				}
				else
				{
					ungetc(ch, pf);
					
					int nTokenType = GetKeywordTokenType(token.szValue);

					if (nTokenType != TT_INVALID)
						return nTokenType;

					return TT_IDENTIFIER;
				}

				break;
			}

			case 2:
			{
				if (ch == '\"')
					return TT_STRING;

				*pszValue++ = ch;
				*pszValue = 0;
				break;
			}

			case 3:
			{
				if (token.nValue == 0)
				{
					if (ch == 'x' || ch == 'X')
					{
						nState = 4;
						break;
					}
				}
				
				if (isdigit(ch))
				{
					// NOTE:  no overflow check!!
					token.nValue = 10 * token.nValue + ch - '0';
					break;
				}

				ungetc(ch, pf);

				return TT_INTEGER;
			}

			case 4:
			{
				if (isxdigit(ch))
				{
					int v = toupper(ch);

					if (v >= 'A' && v <= 'F')
						v -= 'A' + 10;
					else
						v -= '0';

					token.nValue = 16 * token.nValue + v;
					break;
				}

				ungetc(ch, pf);

				return TT_INTEGER;

				break;
			}

			case 5:
			{
				if (ch == '/')
				{
					nState = 6;
					break;
				}

				if (ch == '*')
				{
					nState = 7;
					break;
				}

				break;
			}

			case 6:
			{
				if (ch != '\n')
					nState = 0;
					
				break;
			}

			case 7:
			{
				if (ch == '*')
					nState = 8;

				break;
			}

			case 8:
			{
				if (ch == '/')
					nState = 0;
				else
					nState = 5;

				break;
			}
		}
	}
}

SCRIPT_GAME _rgGames[256]; // arbitrary limitation
int _cGames;

HRESULT ScriptFindGame(LPCTSTR pszName, SCRIPT_GAME &game)
{
	for (int i = 0; i < _cGames; i++)
	{
		if (lstrcmpi(pszName, _rgGames[i].szName) == 0)
		{
			game = _rgGames[i];
			return S_OK;
		}
	}

	return E_FAIL;
}

static HRESULT FindRomset(const SCRIPT_GAME &game, LPCTSTR pszName, SCRIPT_ROMSET &romset)
{
	for (int i = 0; i < game.cRomsets; i++)
	{
		if (lstrcmpi(pszName, game.rgRomsets[i].szName) == 0)
		{
			romset = game.rgRomsets[i];
			return S_OK;
		}
	}

	return E_FAIL;
}

HRESULT ScriptLoadRoms(const SCRIPT_ROMSET &romset, LPCTSTR pszPath, LPBYTE rgbBase)
{
	HRESULT hr = S_OK;

	for (int i = 0; i < romset.cEntries; i++)
	{
		const SCRIPT_ROMSET::ENTRY &entry = romset.rgEntries[i];

		FILE *pf = fopen(TStr<_MAX_PATH>("%s\\%s", pszPath, entry.szFilename), "rb");

		if (pf != NULL)
		{
			if (fseek(pf, entry.dwFileOffset, SEEK_SET) != 0)
			{
				hr = E_FAIL;
			}
			else if (fread(rgbBase + entry.dwLoadOffset, 1, entry.dwLoadLength, pf) != entry.dwLoadLength)
			{
				hr = E_FAIL;
			}

			fclose(pf);

			if (FAILED(hr))
				break;
		}
		else
		{
			hr = E_FAIL;
		}
	}

	return hr;
}

HRESULT ScriptLoadRoms(const SCRIPT_GAME &game, LPCTSTR pszRomsetName, LPBYTE rgbBase)
{
	SCRIPT_ROMSET romset;

	if (FAILED(FindRomset(game, pszRomsetName, romset)))
		return E_FAIL;

	return ScriptLoadRoms(romset, game.szPath, rgbBase);
}

static HWND _hwndOwner = NULL;

extern "C" void ScriptError(LPCTSTR pszFormat, ...)
{
	TCHAR szBuf[512];
	
	va_list marker;
	va_start(marker, pszFormat);
	
	_vsnprintf(szBuf, DIM(szBuf), pszFormat, marker);

	MessageBox(_hwndOwner, szBuf, "Error Processing Script", MB_ICONERROR);
}

BOOL Match(FILE *pf, TOKEN &token, int nTokenType)
{
	int ntt = GetToken(pf, token);
	
	if (ntt == nTokenType)
		return TRUE;
		
	ScriptError("line %i: unexpected token type %s: expected %s\n", _lineno,
		_pszTokenTypeNames[ntt], _pszTokenTypeNames[nTokenType]);

	return FALSE;
}

BOOL ProcessRomset(FILE *pf, SCRIPT_GAME &game)
{
	TOKEN token;
	SCRIPT_ROMSET &romset = game.rgRomsets[game.cRomsets];

	ZERO_INIT(romset);

	if (!Match(pf, token, TT_IDENTIFIER))
		return FALSE;

	lstrcpy(romset.szName, token.szValue);

	if (!Match(pf, token, TT_LBRACE))
		return FALSE;

	int stat = 0;

	for (;;)
	{
		int nTokenType;
		
		if (stat == 0)
			nTokenType = GetToken(pf, token);
		
		switch (nTokenType)
		{
			case TT_ROM:
			{
				if (!Match(pf, token, TT_STRING))
					return FALSE;

				SCRIPT_ROMSET::ENTRY &entry = romset.rgEntries[romset.cEntries];

				lstrcpy(entry.szFilename, token.szValue);

				if (GetToken(pf, token) == TT_COMMA)
				{
					if (!Match(pf, token, TT_INTEGER))
						return FALSE;

					entry.dwLoadOffset = token.nValue;

					if (GetToken(pf, token) == TT_COMMA)
					{
						if (!Match(pf, token, TT_INTEGER))
							return FALSE;

						entry.dwLoadLength = token.nValue;

						if ((nTokenType = GetToken(pf, token)) == TT_COMMA)
						{
							if (!Match(pf, token, TT_INTEGER))
								return FALSE;

							entry.dwFileOffset = token.nValue;
						}
						else
						{
							stat = 1;
						}

						romset.cEntries++;
					}
				}

				break;
			}

			case TT_RBRACE:
			{
				game.cRomsets++;
				return TRUE;
			}

			default:
			{
				ScriptError("line %i: unexpected token type %s\n", _lineno,
					_pszTokenTypeNames[nTokenType]);

				return FALSE;
			}
		}
	}
}

BOOL ProcessGame(FILE *pf, SCRIPT_GAME &game)
{
	TOKEN token;

	ZERO_INIT(game);
	
	if (!Match(pf, token, TT_IDENTIFIER))
		return FALSE;

	lstrcpy(game.szName, token.szValue);

	if (!Match(pf, token, TT_LBRACE))
		return FALSE;

	for (;;)
	{
		int nTokenType;
		
		switch (nTokenType = GetToken(pf, token))
		{
			case TT_TITLE:
			{
				if (!Match(pf, token, TT_STRING))
					return FALSE;

				lstrcpy(game.szTitle, token.szValue);

				break;
			}

			case TT_PATH:
			{
				if (!Match(pf, token, TT_STRING))
					return FALSE;

				lstrcpy(game.szPath, token.szValue);

				break;
			}

			case TT_DRIVER:
			{
				if (!Match(pf, token, TT_IDENTIFIER))
					return FALSE;

				lstrcpy(game.szDriver, token.szValue);
				
				break;
			}

			case TT_ROMSET:
			{
				if (!ProcessRomset(pf, game))
					return FALSE;

				break;
			}

			case TT_RBRACE:
			{
				return TRUE;
			}

			default:
			{
				ScriptError("line %i: unexpected token type %s\n", _lineno,
					_pszTokenTypeNames[nTokenType]);

				return FALSE;
			}
		}
	}
}

HRESULT ScriptInit(LPCTSTR pszFilename)
{
	FILE *pf = fopen(pszFilename, "ra");

	HRESULT hr = S_OK;

	if (pf != NULL)
	{
		TOKEN token;

/*
		while ((i = GetToken(pf, token)) != TT_INVALID)
			ScriptError("token %i: %s, %i\n", i, token.szValue, token.nValue);
*/

		for (;;)
		{
			int nTokenType = GetToken(pf, token);
			
			if (nTokenType == TT_GAME)
			{
				SCRIPT_GAME game;

				TRACE("processing game %i...\n", _cGames + 1);
				
				if (ProcessGame(pf, game))
				{
					if (_cGames < DIM(_rgGames))
					{
						_rgGames[_cGames++] = game;

						TRACE("    name:   %s\n", game.szName);
						TRACE("    title:  %s\n", game.szTitle);
						TRACE("    path:   %s\n", game.szPath);
						TRACE("    driver: %s\n", game.szDriver);

						for (int i = 0; i < game.cRomsets; i++)
						{
							const SCRIPT_ROMSET &romset = game.rgRomsets[i];

							TRACE("    romset #%i - %s\n", i + 1, romset.szName);

							for (int j = 0; j < romset.cEntries; j++)
							{
								const SCRIPT_ROMSET::ENTRY &entry = romset.rgEntries[j];

								TRACE("        rom %04X %04X %04X %s\n",
									entry.dwLoadLength, entry.dwLoadOffset,
									entry.dwFileOffset, entry.szFilename);
							}
						}
					}
					else
					{
						ScriptError("too many games\n");
						break;
					}
				}
				else
				{
					ScriptError("processing game %i failed\n", _cGames + 1);
					break;
				}
			}
			else if (nTokenType == TT_NONE)
			{
				break;
			}
			else
			{
				ScriptError("line %i: unexpected token type %s\n", _lineno,
					_pszTokenTypeNames[nTokenType]);

				break;
			}
		}

		fclose(pf);
	}

	return hr;
}

HRESULT ScriptTerm()
{
	return S_OK;
}

