/*---------------------------------------------------------------------------

	File:
		execz80.cpp

	Abstract:
		Internal declarations for Zilog Z80 emulator.

	Revisions:
		18-Apr-97 EAM Created

	Copyright (C) 1997 Edward Massey

	Optimizations:
	- inlined cCycles -= x;
	- no call table
	- effective PC and SP stored in m_rgbOpcode, m_rgbStack (maybe we can
		do the same thing with HL, IX, and IY?)
	- flags not necessarily stored internally in Z80 format
	- direct access to memory regions where indicated by user

	NOTE:  There is a speed hack for Galaga that I got from MAME (see jr,
		jp) which will be replaced by a cleaner mechanism which dosen't
		affect the Z80 emulation.

---------------------------------------------------------------------------*/

#include "pch.h"
#include "execz80i.h"

extern "C"
{
	int Z80_Dasm (unsigned char *buffer,char *dest,unsigned PC);
};

inline void swap(WORD &b1, WORD &b2)
{
	WORD b = b1;
	b1 = b2;
	b2 = b;
}

#ifdef _DEBUG
extern "C" extern int _rgcIns[];
extern "C" void TRACE(LPCTSTR pszFormat, ...);
#endif

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

static void OutStub(BYTE bPort, BYTE bValue)
{
}

static BYTE InStub(BYTE bPort)
{
	return 0;
}

static BYTE ReadIllegal(WORD wAddr)
{
#ifdef _DEBUG
	TRACE("Z80: illegal read from address %04X\n", wAddr);
#endif
	
	return 0;
}

static void WriteIllegal(WORD wAddr, BYTE bVal)
{
#ifdef _DEBUG
	TRACE("Z80:  illegal write to address %04X (value=%02X)\n", wAddr, bVal);
#endif
}

static BYTE ReadIgnore(WORD wAddr)
{
	return 0;
}

static void WriteIgnore(WORD wAddr, BYTE bVal)
{
}


/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

HRESULT CExecZ80::Create(EXECZ80_CREATE &cs, IExecZ80 *&pExec)
{
 	pExec = new CExecZ80(cs);

	if (pExec == NULL)
		return E_OUTOFMEMORY;

	pExec->AddRef();

	return S_OK;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

CExecZ80::CExecZ80(EXECZ80_CREATE &cs)
{
	//
	//  Initialize pointers
	//
	
	m_rgbMemory = cs.rgbMemory;
	m_rgbOpcodeBase = cs.rgbInstructions ? cs.rgbInstructions : cs.rgbMemory;
	m_rgbStackBase = m_rgbMemory;

	m_cRef = 0;

	m_fPendingInterrupt = FALSE;

	//
	//  Initialize memory address descriptor tables
	//

	for (DWORD dwAddr = 0; dwAddr < 0x10000; dwAddr++)
	{
		EXECZ80_MEMREAD_PROC pfnMemRead = ReadIllegal;
		EXECZ80_MEMWRITE_PROC pfnMemWrite = WriteIllegal;
		
		const EXECZ80_MEM_DESCRIPTOR *pmd;

		for (pmd = cs.pmd; pmd->pfnMemRead != Z80_END_TABLE; pmd++)
		{
			if (dwAddr < pmd->dwAddrMin || dwAddr > pmd->dwAddrMax)
				continue;

			pfnMemRead  = pmd->pfnMemRead;
			pfnMemWrite = pmd->pfnMemWrite;

			if (pfnMemRead == Z80_IGNORE_READ)
				pfnMemRead = ReadIgnore;
			
			if (pfnMemRead == Z80_ILLEGAL_READ)
				pfnMemRead = ReadIllegal;

			if (pfnMemWrite == Z80_IGNORE_WRITE)
				pfnMemWrite = WriteIgnore;
			
			if (pfnMemWrite == Z80_ILLEGAL_WRITE)
				pfnMemWrite = WriteIllegal;
		}

		m_rgpfnMemRead[dwAddr]  = pfnMemRead;
		m_rgpfnMemWrite[dwAddr] = pfnMemWrite;
	}

	//
	//
	//
	
	for (int nPort = 0; nPort < 256; nPort++)
	{
		m_rgpfnPortRead[nPort]  = cs.pfnPortRead  ? cs.pfnPortRead  : InStub;
		m_rgpfnPortWrite[nPort] = cs.pfnPortWrite ? cs.pfnPortWrite : OutStub;
	}

	//
	//
	//
	
	Dump();

	//
	//
	//
	
	Reset();
}

CExecZ80::~CExecZ80()
{
	_ASSERT(m_cRef == 0);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

HRESULT CExecZ80::QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject)
{
	if (riid == IID_IUnknown)
		*ppvObject = this;
	else
		return E_NOINTERFACE;
	
	AddRef();
	
	return S_OK;
}

ULONG CExecZ80::AddRef()
{
	return ++m_cRef;
}

ULONG CExecZ80::Release()
{
	_ASSERT(m_cRef > 0);

	if (--m_cRef != 0)
		return m_cRef;

	delete this;

	return 0;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

inline void CExecZ80::Rlca()
{
	m_regA = (m_regA << 1) | ((m_regA & 0x80) >> 7);
	m_regF = (m_regF & 0xEC) | (m_regA & C_FLAG);
}

inline void CExecZ80::Rrca()
{
	m_regF=(m_regF&0xEC)|(m_regA&0x01);
	m_regA=(m_regA>>1)|(m_regA<<7);
}

inline void CExecZ80::Rla()
{
	int i  = m_regF&C_FLAG;
	m_regF = (m_regF&0xEC)|((m_regA&0x80)>>7);
	m_regA = (m_regA<<1)|i;
}

inline void CExecZ80::Rra()
{
	int i = m_regF&C_FLAG;
	m_regF=(m_regF&0xEC)|(m_regA&0x01);
	m_regA=(m_regA>>1)|(i<<7);
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

#ifdef _DEBUG

struct FXN_NAME
{
	LPCTSTR pszName;
	EXECZ80_MEMREAD_PROC pfnMemRead;
	EXECZ80_MEMWRITE_PROC pfnMemWrite;
};

static const FXN_NAME _rgFxnName[] =
{
	{ "direct    ", NULL,        NULL         },
	{ "ignore    ", ReadIgnore,  WriteIgnore  },
	{ "illegal   ", ReadIllegal, WriteIllegal },
};

void CExecZ80::Dump()
{
	DWORD dwAddrMin = 0;

	TRACE("Z80 memory descriptor dump (this=%p)\n", this);
	TRACE("\n");
	TRACE("addresses   read       write\n");
	TRACE("---------   ---------- ----------\n");

	for (DWORD dwAddr = 0; dwAddr < 0x10000; dwAddr++)
	{
		if ((m_rgpfnMemRead[dwAddr]  != m_rgpfnMemRead[dwAddrMin]) ||
			(m_rgpfnMemWrite[dwAddr] != m_rgpfnMemWrite[dwAddrMin]) ||
			dwAddr == 0xFFFF)
		{
			LPCTSTR pszRead, pszWrite;
			
			pszRead = pszWrite = "callback  ";

			for (int i = 0; i < 3; i++)
			{
				if (m_rgpfnMemRead[dwAddrMin] == _rgFxnName[i].pfnMemRead)
					pszRead = _rgFxnName[i].pszName;

				if (m_rgpfnMemWrite[dwAddrMin] == _rgFxnName[i].pfnMemWrite)
					pszWrite = _rgFxnName[i].pszName;
			}

			TRACE("%04X-%04X   %s %s\n", dwAddrMin, dwAddr - 1,
				pszRead, pszWrite);
				
			dwAddrMin = dwAddr;
		}
	}

	TRACE("\n");
}

#endif

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

void CExecZ80::GetContext(EXECZ80_CONTEXT &ctx)
{
	ctx.AF = m_regAF;
	ctx.BC = m_regBC;
	ctx.DE = m_regDE;
	ctx.HL = m_regHL;
	ctx.IX = m_regIX;
	ctx.IY = m_regIY;
	ctx.PC = GetPC();
	ctx.SP = GetSP();

	ctx.AF2 = m_regAF2;
	ctx.BC2 = m_regBC2;
	ctx.DE2 = m_regDE2;
	ctx.HL2 = m_regHL2;

	ctx.IFF1 = m_iff1;
	ctx.IFF2 = m_iff2;
	ctx.HALT = m_fHalt;
	ctx.IM = m_nIM;
	ctx.I = m_regI;
	ctx.R = m_regR;
	ctx.R2 = 0;
}

void CExecZ80::SetContext(const EXECZ80_CONTEXT &ctx)
{
	m_regAF = ctx.AF;
	m_regBC = ctx.BC;
	m_regDE = ctx.DE;
	m_regHL = ctx.HL;
	m_regIX = ctx.IX;
	m_regIY = ctx.IY;
	SetPC(ctx.PC);
	SetSP(ctx.SP);

	m_regAF2 = ctx.AF2;
	m_regBC2 = ctx.BC2;
	m_regDE2 = ctx.DE2;
	m_regHL2 = ctx.HL2;

	m_iff1 = ctx.IFF1;
	m_iff2 = ctx.IFF2;
	m_fHalt = ctx.HALT;
	m_nIM = ctx.IM;
	m_regI = ctx.I;
	m_regR = ctx.R;
}

/*---------------------------------------------------------------------------
---------------------------------------------------------------------------*/

int CExecZ80::Ei()
{
	m_iff2 = 1;

	if (m_iff1 != 0)
		return 0;

	m_iff1 = 1;

	m_regR++;

	int cCycles = Exec(1);

//	cCycles += Irq(0xFF);

	return cCycles;
}

void CExecZ80::Di()
{
	m_iff1 = 0;
	m_iff2 = 0;
}

/*---------------------------------------------------------------------------

	Flow of control operations

---------------------------------------------------------------------------*/

inline int CExecZ80::Jr0(int f)
{
	if (f)
	{
		AdjustPC((signed char)ImmedByte());
		return 5;
	}
	else
	{
		AdjustPC(1);
		return 0;
	}
}

inline int CExecZ80::Jr1(int f)
{
	if (f)
	{
		AdjustPC(1);
		return 0;
	}
	else
	{
		AdjustPC((signed char)ImmedByte());
		return 5;
	}
}

inline int CExecZ80::Call0(int f)
{
	if (f)
	{
		WORD wAddr = ImmedWord();
		Push(GetPC());
		SetPC(wAddr);
		return 7;
	}
	else
	{
		AdjustPC(2);
		return 0;
	}
}

inline int CExecZ80::Call1(int f)
{
	if (f)
	{
		AdjustPC(2);
		return 0;
	}
	else
	{
		WORD wAddr = ImmedWord();
		Push(GetPC());
		SetPC(wAddr);
		return 7;
	}
}

inline int CExecZ80::Jp0(int f)
{
	if (f)
	{
		SetPC(ImmedWord());
		return 0; // ????????????????
	}
	else
	{
		AdjustPC(2);
		return 0;
	}
}

inline int CExecZ80::Jp1(int f)
{
	if (f)
	{
		AdjustPC(2);
		return 0;
	}
	else
	{
		SetPC(ImmedWord());
		return 0; // ????????????????
	}
}

inline void CExecZ80::Rst(WORD wAddr)
{
	Push(GetPC());
	SetPC(wAddr);
}

/*---------------------------------------------------------------------------

	Exec

	Executes the specified number of t-states.  Returns the actual number of
	t-states executed.  If a halt instruction is encountered, the execution
	loop exits.  If the processor is currently in a halted state this
	function immediately returns 0.

---------------------------------------------------------------------------*/

int CExecZ80::Exec(int cCyclesArg)
{
	//
	//
	//
	
//	if (m_fHalt)
//		return 0;

	//
	//  Save t-state count for return value calculation
	//
	
	int cCycles = cCyclesArg;

	//
	//
	//
	
	while (cCycles > 0)
	{
#ifdef _DEBUG
		static cins = 0;

		cins++;

		if (cins >= _rgcIns[0] && cins <= _rgcIns[1]) 
		{
			TCHAR szBuf[128];
			Z80_Dasm(&m_rgbMemory[GetPC()], szBuf, GetPC());
			TRACE("%04X %04X %04X %04X %04X %04X %04X %04X %04X %s\n",
				GetPC(), m_regAF, m_regBC, m_regDE, m_regHL, m_regIX, m_regIY, GetSP(), cCycles, szBuf);
		}

		if (cins % 1000 == 0 && GetAsyncKeyState(VK_SHIFT) < 0)
			TRACE("%i\n", cins);
#endif
		const BYTE bOpcode = ImmedByte();

		m_regR++;

		switch (bOpcode)
		{
			case 0x00: // nop     
				cCycles -= 4;
				break;
			
			case 0x01: // ld_bc_word
				cCycles -= 10;
				m_regBC = ImmedWord();
				break;
			
			case 0x02: // ld_xbc_a   
				cCycles -= 7;
				MemWriteByte(m_regBC, m_regA);
				break;
			
			case 0x03: // inc_bc    
				cCycles -= 6;
				m_regBC++;
				break;
			
			case 0x04: // inc_b   
				cCycles -= 4;
				m_regB = Inc(m_regB);
				break;
			
			case 0x05: // dec_b   
				cCycles -= 4;
				m_regB = Dec(m_regB);
				break;
			
			case 0x06: // ld_b_byte  
				cCycles -= 7;
				m_regB = ImmedByte();
				break;
			
			case 0x07: // rlca    
				cCycles -= 4;
				Rlca();
				break;

			case 0x08: // ex_af_af
				cCycles -= 4;
				swap(m_regAF, m_regAF2);
				break;
			
			case 0x09: // add_hl_bc 
				cCycles -= 11;
				m_regHL = Add(m_regHL, m_regBC);
				break;
			
			case 0x0A: // ld_a_xbc   
				cCycles -= 7;
				m_regA = MemReadByte(m_regBC);
				break;
			
			case 0x0B: // dec_bc    
				cCycles -= 6;
				m_regBC--;
				break;
			
			case 0x0C: // inc_c   
				cCycles -= 4;
				m_regC = Inc(m_regC);
				break;
			
			case 0x0D: // dec_c   
				cCycles -= 4;
				m_regC = Dec(m_regC);
				break;
			
			case 0x0E: // ld_c_byte  
				cCycles -= 7;
				m_regC = ImmedByte();
				break;
			
			case 0x0F: // rrca    
				cCycles -= 4;
				Rrca();
				break;

			case 0x10: // djnz
				cCycles -= 8;
				cCycles -= Jr0(--m_regB);
				break;
			
			case 0x11: // ld_de_word
				cCycles -= 10;
				m_regDE = ImmedWord();
				break;
			
			case 0x12: // ld_xde_a   
				cCycles -= 7;
				MemWriteByte(m_regDE, m_regA);
				break;
			
			case 0x13: // inc_de    
				cCycles -= 6;
				m_regDE++;
				break;
			
			case 0x14: // inc_d   
				cCycles -= 4;
				m_regD = Inc(m_regD);
				break;
			
			case 0x15: // dec_d   
				cCycles -= 4;
				m_regD = Dec(m_regD);
				break;
			
			case 0x16: // ld_d_byte  
				cCycles -= 7;
				m_regD = ImmedByte();
				break;
			
			case 0x17: // rla     
				cCycles -= 4;
				Rla();
				break;

			case 0x18: // jr
			{
				cCycles -= 7;
#if 0
				cCycles -= Jr0(TRUE);
#else
				signed char offset = (signed char)ImmedByte();

				cCycles -= 5;

				AdjustPC(offset);

				if (offset == -2)
				{
//					TRACE(__FILE__ ":jr (infinite loop) at %04X\n", GetPC());
					return (cCyclesArg - cCycles);
				}
#endif
				break;
			}
			
			case 0x19: // add_hl_de 
				cCycles -= 11;
				m_regHL = Add(m_regHL, m_regDE);
				break;
			
			case 0x1A: // ld_a_xde   
				cCycles -= 7;
				m_regA = MemReadByte(m_regDE);
				break;
			
			case 0x1B: // dec_de    
				cCycles -= 6;
				m_regDE--;
				break;
			
			case 0x1C: // inc_e   
				cCycles -= 4;
				m_regE = Inc(m_regE); 
				break;
			
			case 0x1D: // dec_e   
				cCycles -= 4;
				m_regE = Dec(m_regE);
				break;
			
			case 0x1E: // ld_e_byte  
				cCycles -= 7;
				m_regE = ImmedByte();
				break;
			
			case 0x1F: // rra     
				cCycles -= 4;
				Rra();
				break;

			case 0x20: // jr_nz   
				cCycles -= 7;
				cCycles -= Jr1(m_regF & Z_FLAG);
				break;
			
			case 0x21: // ld_hl_word
				cCycles -= 10;
				m_regHL = ImmedWord();
				break;
			
			case 0x22: // ld_xword_hl
				cCycles -= 16;
				MemWriteWord(ImmedWord(), m_regHL);
				break;
			
			case 0x23: // inc_hl    
				cCycles -= 6;
				m_regHL++;
				break;
			
			case 0x24: // inc_h   
				cCycles -= 4;
				m_regH = Inc(m_regH);
				break;
			
			case 0x25: // dec_h   
				cCycles -= 4;
				m_regH = Dec(m_regH);
				break;
			
			case 0x26: // ld_h_byte  
				cCycles -= 7;
				m_regH = ImmedByte();
				break;
			
			case 0x27: // daa
			{
				cCycles -= 4;
				int i = m_regA;
				if (m_regF&C_FLAG) i|=256;
				if (m_regF&H_FLAG) i|=512;
				if (m_regF&N_FLAG) i|=1024;
				m_regAF = DAATable[i];
				break;
			}

			case 0x28: // jr_z    
				cCycles -= 7;
				cCycles -= Jr0(m_regF & Z_FLAG);
				break;
			
			case 0x29: // add_hl_hl
				cCycles -= 11;
				m_regHL = Add(m_regHL, m_regHL);
				break;
			
			case 0x2A: // ld_hl_xword
				cCycles -= 16;
				m_regHL = MemReadWord(ImmedWord());
				break;
			
			case 0x2B: // dec_hl    
				cCycles -= 6;
				m_regHL--;
				break;
			
			case 0x2C: // inc_l   
				cCycles -= 4;
				m_regL = Inc(m_regL);
				break;
			
			case 0x2D: // dec_l   
				cCycles -= 4;
				m_regL = Dec(m_regL);
				break;
			
			case 0x2E: // ld_l_byte  
				cCycles -= 7;
				m_regL = ImmedByte();
				break;
			
			case 0x2F: // cpl     
				cCycles -= 4;
				m_regA ^= 0xFF;
				m_regF |= (H_FLAG | N_FLAG);
				break;

			case 0x30: // jr_nc   
				cCycles -= 7;
				cCycles -= Jr1(m_regF & C_FLAG);
				break;
			
			case 0x31: // ld_sp_word
				cCycles -= 10;
				SetSP(ImmedWord());
				break;
			
			case 0x32: // ld_xbyte_a
				cCycles -= 13;
				MemWriteByte(ImmedWord(), m_regA);
				break;
			
			case 0x33: // inc_sp
				cCycles -= 6;
				SetSP(GetSP() + 1);
				break;
			
			case 0x34: // inc_xhl
				cCycles -= 11;
				MemWriteByte(m_regHL, Inc(MemReadByte(m_regHL)));
				break;
			
			case 0x35: // dec_xhl
				cCycles -= 11;
				MemWriteByte(m_regHL, Dec(MemReadByte(m_regHL)));
				break;
			
			case 0x36: // ld_xhl_byte
				cCycles -= 10;
				MemWriteByte(m_regHL, ImmedByte());
				break;
			
			case 0x37: // scf
				cCycles -= 4;
				m_regF = (m_regF & 0xEC) | C_FLAG;
				break;

			case 0x38: // jr_c
				cCycles -= 7;
				cCycles -= Jr0(m_regF & C_FLAG);
				break;
			
			case 0x39: // add_hl_sp
				cCycles -= 11;
				m_regHL = Add(m_regHL, GetSP());
				break;
			
			case 0x3A: // ld_a_xbyte 
				cCycles -= 13;
				m_regA = MemReadByte(ImmedWord());
				break;
			
			case 0x3B: // dec_sp
				cCycles -= 6;
				SetSP(GetSP() - 1);
				break;
			
			case 0x3C: // inc_a
				cCycles -= 4;
				m_regA = Inc(m_regA);
				break;
			
			case 0x3D: // dec_a
				cCycles -= 4;
				m_regA = Dec(m_regA);
				break;
			
			case 0x3E: // ld_a_byte
				cCycles -= 7;
				m_regA = ImmedByte();
				break;
			
			case 0x3F: // ccf
				cCycles -= 4;
				m_regF =((m_regF & 0xED)|((m_regF & 1)<<4))^1;
				break;

			case 0x40: // ld_b_b  
				cCycles -= 4;
				m_regB = m_regB;
				break;
			
			case 0x41: // ld_b_c
				cCycles -= 4;
				m_regB = m_regC;
				break;
			
			case 0x42: // ld_b_d
				cCycles -= 4;
				m_regB = m_regD;
				break;
			
			case 0x43: // ld_b_e
				cCycles -= 4;
				m_regB = m_regE;
				break;
			
			case 0x44: // ld_b_h  
				cCycles -= 4;
				m_regB = m_regH;
				break;
			
			case 0x45: // ld_b_l  
				cCycles -= 4;
				m_regB = m_regL;
				break;
			
			case 0x46: // ld_b_xhl   
				cCycles -= 7;
				m_regB = MemReadByte(m_regHL);
				break;
			
			case 0x47: // ld_b_a  
				cCycles -= 4;
				m_regB = m_regA;
				break;

			case 0x48: // ld_c_b  
				cCycles -= 4;
				m_regC = m_regB;
				break;
			
			case 0x49: // ld_c_c    
				cCycles -= 4;
				m_regC = m_regC;
				break;
			
			case 0x4A: // ld_c_d     
				cCycles -= 4;
				m_regC = m_regD;
				break;
			
			case 0x4B: // ld_c_e    
				cCycles -= 4;
				m_regC = m_regE;
				break;
			
			case 0x4C: // ld_c_h  
				cCycles -= 4;
				m_regC = m_regH;
				break;
			
			case 0x4D: // ld_c_l  
				cCycles -= 4;
				m_regC = m_regL;
				break;
			
			case 0x4E: // ld_c_xhl   
				cCycles -= 7;
				m_regC = MemReadByte(m_regHL);
				break;
			
			case 0x4F: // ld_c_a  
				cCycles -= 4;
				m_regC = m_regA;
				break;

			case 0x50: // ld_d_b  
				cCycles -= 4;
				m_regD = m_regB;
				break;
			
			case 0x51: // ld_d_c    
				cCycles -= 4;
				m_regD = m_regC;
				break;
			
			case 0x52: // ld_d_d     
				cCycles -= 4;
				m_regD = m_regD;
				break;
			
			case 0x53: // ld_d_e    
				cCycles -= 4;
				m_regD = m_regE;
				break;
			
			case 0x54: // ld_d_h  
				cCycles -= 4;
				m_regD = m_regH;
				break;
			
			case 0x55: // ld_d_l  
				cCycles -= 4;
				m_regD = m_regL;
				break;
			
			case 0x56: // ld_d_xhl   
				cCycles -= 7;
				m_regD = MemReadByte(m_regHL);
				break;
			
			case 0x57: // ld_d_a  
				cCycles -= 4;
				m_regD = m_regA;
				break;

			case 0x58: // ld_e_b  
				cCycles -= 4;
				m_regE = m_regB;
				break;
			
			case 0x59: // ld_e_c    
				cCycles -= 4;
				m_regE = m_regC;
				break;
			
			case 0x5A: // ld_e_d     
				cCycles -= 4;
				m_regE = m_regD;
				break;
			
			case 0x5B: // ld_e_e    
				cCycles -= 4;
				m_regE = m_regE;
				break;
			
			case 0x5C: // ld_e_h  
				cCycles -= 4;
				m_regE = m_regH;
				break;
			
			case 0x5D: // ld_e_l  
				cCycles -= 4;
				m_regE = m_regL;
				break;
			
			case 0x5E: // ld_e_xhl   
				cCycles -= 7;
				m_regE = MemReadByte(m_regHL);
				break;
			
			case 0x5F: // ld_e_a  
				cCycles -= 4;
				m_regE = m_regA;
				break;

			case 0x60: // ld_h_b  
				cCycles -= 4;
				m_regH = m_regB;
				break;
			
			case 0x61: // ld_h_c    
				cCycles -= 4;
				m_regH = m_regC;
				break;
			
			case 0x62: // ld_h_d     
				cCycles -= 4;
				m_regH = m_regD;
				break;
			
			case 0x63: // ld_h_e    
				cCycles -= 4;
				m_regH = m_regE;
				break;
			
			case 0x64: // ld_h_h  
				cCycles -= 4;
				m_regH = m_regH;
				break;
			
			case 0x65: // ld_h_l  
				cCycles -= 4;
				m_regH = m_regL;
				break;
			
			case 0x66: // ld_h_xhl   
				cCycles -= 7;
				m_regH = MemReadByte(m_regHL);
				break;
			
			case 0x67: // ld_h_a  
				cCycles -= 4;
				m_regH = m_regA;
				break;

			case 0x68: // ld_l_b  
				cCycles -= 4;
				m_regL = m_regB;
				break;
			
			case 0x69: // ld_l_c    
				cCycles -= 4;
				m_regL = m_regC;
				break;
			
			case 0x6A: // ld_l_d     
				cCycles -= 4;
				m_regL = m_regD;
				break;
			
			case 0x6B: // ld_l_e    
				cCycles -= 4;
				m_regL = m_regE;
				break;
			
			case 0x6C: // ld_l_h  
				cCycles -= 4;
				m_regL = m_regH;
				break;
			
			case 0x6D: // ld_l_l  
				cCycles -= 4;
				m_regL = m_regL;
				break;
			
			case 0x6E: // ld_l_xhl   
				cCycles -= 7;
				m_regL = MemReadByte(m_regHL);
				break;
			
			case 0x6F: // ld_l_a  
				cCycles -= 4;
				m_regL = m_regA;
				break;

			case 0x70: // ld_xhl_b
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regB);
				break;
			
			case 0x71: // ld_xhl_c  
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regC);
				break;
			
			case 0x72: // ld_xhl_d   
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regD);
				break;
			
			case 0x73: // ld_xhl_e  
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regE);
				break;
			
			case 0x74: // ld_xhl_h
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regH);
				break;
			
			case 0x75: // ld_xhl_l
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regL);
				break;
			
			case 0x76: // halt
				cCycles -= 4;
				SetPC(GetPC() - 1);
				m_fHalt = TRUE;
				return (cCyclesArg - cCycles);
			
			case 0x77: // ld_xhl_a
				cCycles -= 7;
				MemWriteByte(m_regHL, m_regA);
				break;

			case 0x78: // ld_a_b  
				cCycles -= 4;
				m_regA = m_regB;
				break;
			
			case 0x79: // ld_a_c    
				cCycles -= 4;
				m_regA = m_regC;
				break;
			
			case 0x7A: // ld_a_d     
				cCycles -= 4;
				m_regA = m_regD;
				break;
			
			case 0x7B: // ld_a_e    
				cCycles -= 4;
				m_regA = m_regE;
				break;
			
			case 0x7C: // ld_a_h  
				cCycles -= 4;
				m_regA = m_regH;
				break;
			
			case 0x7D: // ld_a_l  
				cCycles -= 4;
				m_regA = m_regL;
				break;
			
			case 0x7E: // ld_a_xhl   
				cCycles -= 7;
				m_regA = MemReadByte(m_regHL);
				break;
			
			case 0x7F: // ld_a_a  
				cCycles -= 4;
				m_regA = m_regA;
				break;

			case 0x80: // add_a_b 
				cCycles -= 4;
				Add(m_regB);
				break;
			
			case 0x81: // add_a_c   
				cCycles -= 4;
				Add(m_regC);
				break;
			
			case 0x82: // add_a_d    
				cCycles -= 4;
				Add(m_regD);
				break;
			
			case 0x83: // add_a_e   
				cCycles -= 4;
				Add(m_regE);
				break;
			
			case 0x84: // add_a_h 
				cCycles -= 4;
				Add(m_regH);
				break;
			
			case 0x85: // add_a_l 
				cCycles -= 4;
				Add(m_regL);
				break;
			
			case 0x86: // add_a_xhl  
				cCycles -= 7;
				Add(MemReadByte(m_regHL));
				break;
			
			case 0x87: // add_a_a 
				cCycles -= 4;
				Add(m_regA);
				break;

			case 0x88: // adc_a_b 
				cCycles -= 4;
				Adc(m_regB);
				break;
			
			case 0x89: // adc_a_c   
				cCycles -= 4;
				Adc(m_regC);
				break;
			
			case 0x8A: // adc_a_d    
				cCycles -= 4;
				Adc(m_regD);
				break;
			
			case 0x8B: // adc_a_e   
				cCycles -= 4;
				Adc(m_regE);
				break;
			
			case 0x8C: // adc_a_h 
				cCycles -= 4;
				Adc(m_regH);
				break;
			
			case 0x8D: // adc_a_l 
				cCycles -= 4;
				Adc(m_regL);
				break;
			
			case 0x8E: // adc_a_xhl  
				cCycles -= 7;
				Adc(MemReadByte(m_regHL));
				break;
			
			case 0x8F: // adc_a_a 
				cCycles -= 4;
				Adc(m_regA);
				break;

			case 0x90: // sub_b   
				cCycles -= 4;
				Sub(m_regB);
				break;
			
			case 0x91: // sub_c     
				cCycles -= 4;
				Sub(m_regC);
				break;
			
			case 0x92: // sub_d      
				cCycles -= 4;
				Sub(m_regD);
				break;
			
			case 0x93: // sub_e     
				cCycles -= 4;
				Sub(m_regE);
				break;
			
			case 0x94: // sub_h   
				cCycles -= 4;
				Sub(m_regH);
				break;
			
			case 0x95: // sub_l   
				cCycles -= 4;
				Sub(m_regL);
				break;
			
			case 0x96: // sub_xhl    
				cCycles -= 7;
				Sub(MemReadByte(m_regHL));
				break;
			
			case 0x97: // sub_a   
				cCycles -= 4;
				Sub(m_regA);
				break;

			case 0x98: // sbc_a_b 
				cCycles -= 4;
				Sbc(m_regB);
				break;
			
			case 0x99: // sbc_a_c   
				cCycles -= 4;
				Sbc(m_regC);
				break;
			
			case 0x9A: // sbc_a_d    
				cCycles -= 4;
				Sbc(m_regD);
				break;
			
			case 0x9B: // sbc_a_e   
				cCycles -= 4;
				Sbc(m_regE);
				break;
			
			case 0x9C: // sbc_a_h 
				cCycles -= 4;
				Sbc(m_regH);
				break;
			
			case 0x9D: // sbc_a_l 
				cCycles -= 4;
				Sbc(m_regL);
				break;
			
			case 0x9E: // sbc_a_xhl  
				cCycles -= 7;
				Sbc(MemReadByte(m_regHL));
				break;
			
			case 0x9F: // sbc_a_a 
				cCycles -= 4;
				Sbc(m_regA);
				break;

			case 0xA0: // and_b   
				cCycles -= 4;
				And(m_regB);
				break;
			
			case 0xA1: // and_c     
				cCycles -= 4;
				And(m_regC);
				break;
			
			case 0xA2: // and_d      
				cCycles -= 4;
				And(m_regD);
				break;
			
			case 0xA3: // and_e     
				cCycles -= 4;
				And(m_regE);
				break;
			
			case 0xA4: // and_h   
				cCycles -= 4;
				And(m_regH);
				break;
			
			case 0xA5: // and_l   
				cCycles -= 4;
				And(m_regL);
				break;
			
			case 0xA6: // and_xhl    
				cCycles -= 7;
				And(MemReadByte(m_regHL));
				break;
			
			case 0xA7: // and_a   
				cCycles -= 4;
				And(m_regA);
				break;

			case 0xA8: // xor_b   
				cCycles -= 4;
				Xor(m_regB);
				break;
			
			case 0xA9: // xor_c     
				cCycles -= 4;
				Xor(m_regC);
				break;
			
			case 0xAA: // xor_d      
				cCycles -= 4;
				Xor(m_regD);
				break;
			
			case 0xAB: // xor_e     
				cCycles -= 4;
				Xor(m_regE);
				break;
			
			case 0xAC: // xor_h   
				cCycles -= 4;
				Xor(m_regH);
				break;
			
			case 0xAD: // xor_l   
				cCycles -= 4;
				Xor(m_regL);
				break;
			
			case 0xAE: // xor_xhl    
				cCycles -= 7;
				Xor(MemReadByte(m_regHL));
				break;
			
			case 0xAF: // xor_a   
				cCycles -= 4;
				Xor(m_regA);
				break;

			case 0xB0: // or_b    
				cCycles -= 4;
				Or(m_regB);
				break;
			
			case 0xB1: // or_c      
				cCycles -= 4;
				Or(m_regC);
				break;
			
			case 0xB2: // or_d       
				cCycles -= 4;
				Or(m_regD);
				break;
			
			case 0xB3: // or_e      
				cCycles -= 4;
				Or(m_regE);
				break;
			
			case 0xB4: // or_h    
				cCycles -= 4;
				Or(m_regH);
				break;
			
			case 0xB5: // or_l    
				cCycles -= 4;
				Or(m_regL);
				break;
			
			case 0xB6: // or_xhl     
				cCycles -= 7;
				Or(MemReadByte(m_regHL));
				break;
			
			case 0xB7: // or_a    
				cCycles -= 4;
				Or(m_regA);
				break;

			case 0xB8: // cp_b    
				cCycles -= 4;
				Cp(m_regB);
				break;
			
			case 0xB9: // cp_c      
				cCycles -= 4;
				Cp(m_regC);
				break;
			
			case 0xBA: // cp_d       
				cCycles -= 4;
				Cp(m_regD);
				break;
			
			case 0xBB: // cp_e      
				cCycles -= 4;
				Cp(m_regE);
				break;
			
			case 0xBC: // cp_h    
				cCycles -= 4;
				Cp(m_regH);
				break;
			
			case 0xBD: // cp_l    
				cCycles -= 4;
				Cp(m_regL);
				break;
			
			case 0xBE: // cp_xhl     
				cCycles -= 7;
				Cp(MemReadByte(m_regHL));
				break;
			
			case 0xBF: // cp_a    
				cCycles -= 4;
				Cp(m_regA);
				break;

			case 0xC0: // ret_nz 
				cCycles -= 5;
				cCycles -= Ret1(m_regF & Z_FLAG);
				break;
			
			case 0xC1: // pop_bc    
				cCycles -= 10;
				m_regBC = Pop();
				break;
			
			case 0xC2: // jp_nz      
				cCycles -= 10;
				cCycles -= Jp1(m_regF & Z_FLAG);
				break;
			
			case 0xC3: // jp
				cCycles -= 10;
#if 0
				Jp0(TRUE);
#else
				{
					// speedup hack for Galaga adopted from MAME...
					
					WORD w = ImmedWord(), wPc = GetPC();
					
					SetPC(w);

					if (w == wPc)
					{
	//					TRACE(__FILE__ ":jp (infinite loop) at %04X\n", GetPC());
						return (cCyclesArg - cCycles);
					}
					else if (w == wPc - 6)
					{
						if (*m_rgbOpcode == 0x31)
						{
		//					TRACE(__FILE__ ":jp (ld sp,#xxxx) at %04X\n", GetPC());
							return (cCyclesArg - cCycles);
						}
					}
				}
#endif
				break;
			
			case 0xC4: // call_nz 
				cCycles -= 10;
				cCycles -= Call1(m_regF & Z_FLAG);
				break;
			
			case 0xC5: // push_bc
				cCycles -= 11;
				Push(m_regBC);
				break;
			
			case 0xC6: // add_a_byte 
				cCycles -= 7;
				Add(ImmedByte());
				break;
			
			case 0xC7: // rst_00  
				cCycles -= 11;
				Rst(0x00);
				break;

			case 0xC8: // ret_z
				cCycles -= 5;
				cCycles -= Ret0(m_regF & Z_FLAG);
				break;
			
			case 0xC9: // ret
				cCycles -= 4;
				cCycles -= Ret0(TRUE);
				break;
			
			case 0xCA: // jp_z
				cCycles -= 10;
				cCycles -= Jp0(m_regF & Z_FLAG);
				break;
			
			case 0xCB: // cb
				cCycles -= HandleCB();
				break;
			
			case 0xCC: // call_z
				cCycles -= 10;
				cCycles -= Call0(m_regF & Z_FLAG);
				break;
			
			case 0xCD: // call
				cCycles -= 10;
				cCycles -= Call0(TRUE);
				break;
			
			case 0xCE: // adc_a_byte
				cCycles -= 7;
				Adc(ImmedByte());
				break;
			
			case 0xCF: // rst_08
				cCycles -= 11;
				Rst(0x08);
				break;

			case 0xD0: // ret_nc
				cCycles -= 5;
				cCycles -= Ret1(m_regF & C_FLAG);
				break;
			
			case 0xD1: // pop_de
				cCycles -= 10;
				m_regDE = Pop();
				break;
			
			case 0xD2: // jp_nc
				cCycles -= 10;
				cCycles -= Jp1(m_regF & C_FLAG);
				break;
			
			case 0xD3: // out_byte_a
				cCycles -= 11;
				Out(ImmedByte(), m_regA);
				break;
			
			case 0xD4: // call_nc
				cCycles -= 10;
				cCycles -= Call1(m_regF & C_FLAG);
				break;
			
			case 0xD5: // push_de
				cCycles -= 11;
				Push(m_regDE);
				break;
			
			case 0xD6: // sub_byte
				cCycles -= 7;
				Sub(ImmedByte());
				break;
			
			case 0xD7: // rst_10
				cCycles -= 11;
				Rst(0x10);
				break;

			case 0xD8: // ret_c
				cCycles -= 5;
				cCycles -= Ret0(m_regF & C_FLAG);
				break;
			
			case 0xD9: // exx
				cCycles -= 4;
				Exx();
				break;
			
			case 0xDA: // jp_c
				cCycles -= 10;
				cCycles -= Jp0(m_regF & C_FLAG);
				break;
			
			case 0xDB: // in_a_byte 
			{
				cCycles -= 11;
				BYTE bPort = ImmedByte();
				m_regA = m_rgpfnPortRead[bPort](bPort);
				break;
			}
			
			case 0xDC: // call_c
				cCycles -= 10;
				cCycles -= Call0(m_regF & C_FLAG);
				break;
			
			case 0xDD: // dd
				cCycles -= HandleDD();
				break;
			
			case 0xDE: // sbc_a_byte
				cCycles -= 7;
				Sbc(ImmedByte());
				break;
			
			case 0xDF: // rst_18
				cCycles -= 11;
				Rst(0x18);
				break;

			case 0xE0: // ret_po
				cCycles -= 5;
				cCycles -= Ret1(m_regF & V_FLAG);
				break;
			
			case 0xE1: // pop_hl
				cCycles -= 10;
				m_regHL = Pop();
				break;
			
			case 0xE2: // jp_po
				cCycles -= 10;
				cCycles -= Jp1(m_regF & V_FLAG);
				break;
			
			case 0xE3: // ex_xsp_hl
				cCycles -= 19;
			{
				int i = MemReadWord(GetSP());
				MemWriteWord(GetSP(), m_regHL);
				m_regHL = i;
				break;
			}
			
			case 0xE4: // call_po
				cCycles -= 10;
				cCycles -= Call1(m_regF & V_FLAG);
				break;
			
			case 0xE5: // push_hl
				cCycles -= 11;
				Push(m_regHL);
				break;
			
			case 0xE6: // and_byte
				cCycles -= 7;
				And(ImmedByte());
				break;
			
			case 0xE7: // rst_20
				cCycles -= 11;
				Rst(0x20);
				break;

			case 0xE8: // ret_pe
				cCycles -= 5;
				cCycles -= Ret0(m_regF & V_FLAG);
				break;
			
			case 0xE9: // jp_hl
				cCycles -= 4;
				SetPC(m_regHL);
				break;
			
			case 0xEA: // jp_pe
				cCycles -= 10;
				cCycles -= Jp0(m_regF & V_FLAG);
				break;
			
			case 0xEB: // ex_de_hl
				cCycles -= 4;
				swap(m_regDE, m_regHL);
				break;
			
			case 0xEC: // call_pe
				cCycles -= 10;
				cCycles -= Call0(m_regF & V_FLAG);
				break;
			
			case 0xED: // ed
				cCycles -= HandleED(cCycles);
				break;
			
			case 0xEE: // xor_byte
				cCycles -= 7;
				Xor(ImmedByte());
				break;
			
			case 0xEF: // rst_28
				cCycles -= 11;
				Rst(0x28);
				break;

			case 0xF0: // ret_p
				cCycles -= 5;
				cCycles -= Ret1(m_regF & S_FLAG);
				break;
			
			case 0xF1: // pop_af
				cCycles -= 10;
				m_regAF = Pop();
				break;
			
			case 0xF2: // jp_p
				cCycles -= 10;
				cCycles -= Jp1(m_regF & S_FLAG);
				break;
			
			case 0xF3: // di
				cCycles -= 4;
				Di();
				break;
			
			case 0xF4: // call_p
				cCycles -= 10;
				cCycles -= Call1(m_regF & S_FLAG);
				break;
			
			case 0xF5: // push_af
				cCycles -= 11;
				Push(m_regAF);
				break;
			
			case 0xF6: // or_byte
				cCycles -= 7;
				Or(ImmedByte());
				break;
			
			case 0xF7: // rst_30
				cCycles -= 11;
				Rst(0x30);
				break;

			case 0xF8: // ret_m   
				cCycles -= 5;
				cCycles -= Ret0(m_regF & S_FLAG);
				break;
			
			case 0xF9: // ld_sp_hl
				cCycles -= 6;
				SetSP(m_regHL);
				break;
			
			case 0xFA: // jp_m
				cCycles -= 10;
				cCycles -= Jp0(m_regF & S_FLAG);
				break;
			
			case 0xFB: // ei
				cCycles -= 4;
				cCycles -= Ei();
				if (m_fPendingInterrupt)
				{
					return (cCyclesArg - cCycles);
				}
				break;
			
			case 0xFC: // call_m
				cCycles -= 10;
				cCycles -= Call0(m_regF & S_FLAG);
				break;
			
			case 0xFD: // fd
				cCycles -= HandleFD();
				break;
			
			case 0xFE: // cp_byte    
				cCycles -= 7;
				Cp(ImmedByte());
				break;
			
			case 0xFF: // rst_38
				cCycles -= 11;
				Rst(0x38);
				break;

			default:
				_ASSERT(FALSE);
				break;
		}
	}

	return (cCyclesArg - cCycles);
}

int CExecZ80::Irq(BYTE bVal)
{
	if (m_iff1 == 0) // is IRQ disabled?
	{
		m_fPendingInterrupt = TRUE;
		return 0;
	}
		
	m_fPendingInterrupt = FALSE;

	m_iff1 = 0;

	if (m_fHalt)
	{
		SetPC(GetPC() + 1);
		m_fHalt = FALSE;
	}

	switch (m_nIM)
	{
		case 0:
		{
			switch (bVal)
			{
				case 0xCD: // opcode for call
					Push(GetPC());
			
				case 0xC3: // opcode for jump
					SetPC(ImmedWord());
					break;

				case 0xC7: // rst_00  
					Rst(0x00);
					break;

				case 0xCF: // rst_08
					Rst(0x08);
					break;

				case 0xD7: // rst_10
					Rst(0x10);
					break;
				
				case 0xDF: // rst_18
					Rst(0x18);
					break;

				case 0xE7: // rst_20
					Rst(0x20);
					break;

				case 0xEF: // rst_28
					Rst(0x28);
					break;

				case 0xF7: // rst_30
					Rst(0x30);
					break;
				
				case 0xFF: // rst_38
					Rst(0x38);
					break;
				
				default:
					_ASSERT(FALSE); // this code can't handle this condition yet!!!
			}
		}

		case 1:
		{
			Rst(0x38);
			return 11;
		}

		case 2:
		{
			Push(GetPC());
			SetPC(MemReadWord(MAKEWORD(bVal, m_regI)));
			return 0;
		}

		default:
			_ASSERT(FALSE);
	}

	return 0; // fix-me
}

int CExecZ80::Nmi()
{
	m_iff1 = 0;

	if (m_fHalt)
	{
		SetPC(GetPC() + 1);
		m_fHalt = FALSE;
	}
	
	Rst(0x66);

	return 0; // fix-me
}

void CExecZ80::Reset()
{
	m_regAF = 0;
	m_regBC = 0;
	m_regDE = 0;
	m_regHL = 0;
	m_regAF2 = 0;
	m_regBC2 = 0;
	m_regDE2 = 0;
	m_regHL2 = 0;
	m_regR = rand();
	m_regI = 0;
	m_fHalt = FALSE;
	m_nIM = 0;
	m_regIX = 0;
	m_regIY = 0;
	m_iff1 = 0;
	m_iff2 = 0;

	SetPC(0x0000);
	SetSP(0xF000);
}

void CExecZ80::Exx()
{
	swap(m_regBC, m_regBC2);
	swap(m_regDE, m_regDE2);
	swap(m_regHL, m_regHL2);
}

BYTE CExecZ80::In(BYTE bPort)
{
	BYTE bVal = m_rgpfnPortRead[bPort](bPort);
	m_regF = (m_regF & C_FLAG) | ZSPTable[bVal];
	return bVal;
}

void CExecZ80::Out(BYTE bPort, BYTE bVal)
{
	m_rgpfnPortWrite[bPort](bPort, bVal);
}

