/*---------------------------------------------------------------------------

	File:
		execz80i.h

	Abstract:
		Internal declarations for Zilog Z80 emulator.

	Revisions:
		18-Apr-97 EAM Created
		04-Jun-97 EAM 

	Copyright (C) 1997 Edward Massey

	NOTE:  In the current version, the arithmetic operations are directly
		derived from the Portable Z80 Emulator by Marat Fayzullin and
		Marcel de Kogel.

	This library is free software; you can redistribute it and/or modify it
	under the terms of version 2 of the GNU Library General Public License
	as published by the Free Software Foundation.

	This library is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library
	General Public License for more details.
	To obtain a copy of the GNU Library General Public License, write to the
	Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	Any permitted reproduction of these routines, in whole or in part, must
	bear this legend.

---------------------------------------------------------------------------*/

#ifndef _NEWZ80I_H_
#define _NEWZ80I_H_

#define EXECZ80_API extern "C" __declspec(dllexport)

#include "execz80.h"

class CExecZ80 : public IExecZ80
{
public:
	static HRESULT Create(EXECZ80_CREATE &cs, IExecZ80 *&pExec);

	// *** IUnknown methods ***
	virtual HRESULT __stdcall QueryInterface(REFIID riid, void __RPC_FAR *__RPC_FAR *ppvObject);
	virtual ULONG __stdcall AddRef();
	virtual ULONG __stdcall Release();

	// *** IExecZ80 methods ***
	virtual void GetContext(EXECZ80_CONTEXT &ctx);
	virtual void SetContext(const EXECZ80_CONTEXT &ctx);
	virtual void Reset();
	virtual int Exec(int cCycles);
	virtual int Irq(BYTE bVal);
	virtual int Nmi();

protected:
	CExecZ80(EXECZ80_CREATE &cs);
	virtual ~CExecZ80();

	inline BYTE ImmedByte();
	inline WORD ImmedWord();
	inline WORD GetPC();
	inline void SetPC(WORD wAddr);
	inline void AdjustPC(signed char cb);

	inline void Push(WORD wArg);
	inline WORD Pop();
	inline WORD GetSP();
	inline void SetSP(WORD wAddr);
	
	inline int Jr0(int f);
	inline int Jr1(int f);
	inline int Call0(int f);
	inline int Call1(int f);
	inline int Ret0(int f);
	inline int Ret1(int f);
	inline int Jp0(int f);
	inline int Jp1(int f);
	
	inline void Rst(WORD wAddr);
	inline WORD MemReadWord(WORD wAddr);
	inline BYTE MemReadByte(WORD wAddr);
	inline void MemWriteByte(WORD wAddr, BYTE bVal);
	inline void MemWriteWord(WORD wAddr, WORD wVal);
	WORD CallbackMemReadWord(WORD wAddr);
	BYTE CallbackMemReadByte(WORD wAddr);
	void CallbackMemWriteByte(WORD wAddr, BYTE bVal);
	void CallbackMemWriteWord(WORD wAddr, WORD wVal);
	
	inline void Add(BYTE b);
	inline void Adc(BYTE b);
	inline void Sub(BYTE b);
	inline void Sbc(BYTE b);
	inline void And(BYTE b);
	inline void Or(BYTE b);
	inline void Xor(BYTE b);
	inline void Cp(BYTE b);
	inline BYTE Set(BYTE b,int);
	inline BYTE Res(BYTE b,int);
	inline void Bit(BYTE b,int);
	inline void Rlca();
	inline void Rrca();
	inline void Rla();
	inline void Rra();
	inline BYTE Rlc(BYTE b);
	inline BYTE Rrc(BYTE b);
	inline BYTE Rl(BYTE b);
	inline BYTE Rr(BYTE b);
	inline BYTE Sll(BYTE b);
	inline BYTE Srl(BYTE b);
	inline BYTE Sla(BYTE b);
	inline BYTE Sra(BYTE b);
	inline BYTE Inc(BYTE b);
	inline BYTE Dec(BYTE b);

	inline WORD Add(WORD wArg1, WORD wArg2);
	inline WORD Adc(WORD wArg1, WORD wArg2);
	inline WORD Sbc(WORD wArg1, WORD wArg2);

	int Ei();
	void Di();

	int HandleCB();
	int HandleDD();
	int HandleED(int cCycles);
	int HandleFD();
	int HandleDDCB();
	int HandleFDCB();

	void Out(BYTE bPort, BYTE bVal);
	BYTE In(BYTE bPort);

	void Exx();

	inline WORD IndirectIX();
	inline WORD IndirectIY();

	void Dump();

private:
	ULONG m_cRef; // reference count

	//
	//
	//

	union { WORD m_regAF; struct { BYTE m_regF; BYTE m_regA; }; };
	union { WORD m_regBC; struct { BYTE m_regC; BYTE m_regB; }; };
	union { WORD m_regDE; struct { BYTE m_regE; BYTE m_regD; }; };
	union { WORD m_regHL; struct { BYTE m_regL; BYTE m_regH; }; };
	union { WORD m_regIX; struct { BYTE m_regIXl; BYTE m_regIXh; }; };
	union { WORD m_regIY; struct { BYTE m_regIYl; BYTE m_regIYh; }; };

	//
	//
	//
	
	const BYTE *m_rgbOpcode; // takes place of the PC register
	LPBYTE m_rgbStack; // takes place of the SP register
	LPBYTE m_rgbMemory; // direct access to memory buffer (RAM)
	const BYTE *m_rgbOpcodeBase; // "base" pointer for m_rgbOpcode
	LPBYTE m_rgbStackBase;

	BYTE m_regR; // consider:  should be int for performance?
	BYTE m_regI; // consider:  combine into

	// TODO:  combine into a single int...
	int m_iff1, m_iff2;
	BOOL m_fHalt;
	int m_nIM;
	BOOL m_fPendingInterrupt;

//	DWORD m_rgfInterrupt;

	//
	//
	//
	
	EXECZ80_MEMREAD_PROC m_rgpfnMemRead[0x10000];
	EXECZ80_MEMWRITE_PROC m_rgpfnMemWrite[0x10000];

	EXECZ80_PORTREAD_PROC m_rgpfnPortRead[0x100];
	EXECZ80_PORTWRITE_PROC m_rgpfnPortWrite[0x100];

	//
	//
	//
	
	WORD m_regAF2;
	WORD m_regBC2;
	WORD m_regDE2;
	WORD m_regHL2;
};

extern BYTE PTable[512];
extern BYTE ZSTable[512];
extern BYTE ZSPTable[512];
extern short DAATable[2048];
extern BYTE rgfAdd[256][256];
extern BYTE rgfAddc[256][256][2];
extern BYTE rgfSub[256][256];
extern BYTE rgfSubc[256][256][2];

//
//  Memory access flags...
//

#define ISF_PENDING   0x00000008
#define ISF_MODE_MASK 0x00000003
#define ISF_HALT      0x00000004
#define ISF_IFF1      0x00000010
#define ISF_IFF2      0x00000020
#define ISF_IREG_MASK 0x00000F00

/*---------------------------------------------------------------------------

	Basic instruction pointer operations

---------------------------------------------------------------------------*/

inline BYTE CExecZ80::ImmedByte()
{
	return *m_rgbOpcode++;
}

inline WORD CExecZ80::ImmedWord()
{
	return MAKEWORD(*m_rgbOpcode++, *m_rgbOpcode++);
}

inline WORD CExecZ80::GetPC()
{
	return (WORD)(m_rgbOpcode - m_rgbOpcodeBase);
}

inline void CExecZ80::SetPC(WORD wAddr)
{
	m_rgbOpcode = &m_rgbOpcodeBase[wAddr];
}

inline void CExecZ80::AdjustPC(signed char cb)
{
	m_rgbOpcode += cb;
}

/*---------------------------------------------------------------------------

	Memory access

---------------------------------------------------------------------------*/

inline BYTE CExecZ80::MemReadByte(WORD wAddr)
{
	if (m_rgpfnMemRead[wAddr] != NULL)
		return m_rgpfnMemRead[wAddr](wAddr);
	else
		return m_rgbMemory[wAddr];
}

inline WORD CExecZ80::MemReadWord(WORD wAddr)
{
	return MAKEWORD(MemReadByte(wAddr++), MemReadByte(wAddr++));
}

inline void CExecZ80::MemWriteByte(WORD wAddr, BYTE bVal)
{
	if (m_rgpfnMemWrite[wAddr] != NULL)
		m_rgpfnMemWrite[wAddr](wAddr, bVal);
	else
		m_rgbMemory[wAddr] = bVal;
}

inline void CExecZ80::MemWriteWord(WORD wAddr, WORD wVal)
{
	MemWriteByte(wAddr++, LOBYTE(wVal));
	MemWriteByte(wAddr,   HIBYTE(wVal));
}

/*---------------------------------------------------------------------------

	Stack operations

	add_ix_sp/add_iy_sp
	ex_xsp_ix/ex_xsp_iy
	ld_sp_ix/ld_sp_iy
	ld_sp_word
	inc_sp/dec_sp
	add_hl_sp
	ex_xsp_hl
	ld_sp_hl

---------------------------------------------------------------------------*/

#if 0

inline void CExecZ80::Push(WORD wArg)
{
	*--m_rgbStack = HIBYTE(wArg);
	*--m_rgbStack = LOBYTE(wArg);
}

inline WORD CExecZ80::Pop()
{
	return MAKEWORD(*m_rgbStack++, *m_rgbStack++);
}

inline WORD CExecZ80::GetSP()
{
	return (WORD)(m_rgbStack - m_rgbStackBase);
}

inline void CExecZ80::SetSP(WORD wAddr)
{
	m_rgbStack = &m_rgbStackBase[wAddr];
}

#else

inline void CExecZ80::Push(WORD wArg)
{
	MemWriteByte(--m_rgbStack - m_rgbStackBase, HIBYTE(wArg));
	MemWriteByte(--m_rgbStack - m_rgbStackBase, LOBYTE(wArg));
}

inline WORD CExecZ80::Pop()
{
	return MAKEWORD(MemReadByte(m_rgbStack++ - m_rgbStackBase), MemReadByte(m_rgbStack++ - m_rgbStackBase));
}

inline WORD CExecZ80::GetSP()
{
	return (WORD)(m_rgbStack - m_rgbStackBase);
}

inline void CExecZ80::SetSP(WORD wAddr)
{
	m_rgbStack = &m_rgbStackBase[wAddr];
}

#endif

/*---------------------------------------------------------------------------

	Flow of control operations

---------------------------------------------------------------------------*/

inline int CExecZ80::Ret0(int f)
{
	if (f)
	{
		SetPC(Pop());
		return 6;
	}
	else
	{
		return 0;
	}
}

inline int CExecZ80::Ret1(int f)
{
	if (f)
	{
		return 0;
	}
	else
	{
		SetPC(Pop());
		return 6;
	}
}

/*---------------------------------------------------------------------------

	Arithmetic operations

---------------------------------------------------------------------------*/

//
//  16-bit
//

inline WORD CExecZ80::Adc(WORD wArg1, WORD wArg2)
{
	int q  = wArg1 + wArg2 + (m_regF & C_FLAG);
	m_regF = (((wArg1^q^wArg2)&0x1000)>>8)|((q>>16)&1)|((q&0x8000)>>8)|((q&65535)?0:Z_FLAG)|(((wArg2^wArg1^0x8000)&(wArg2^q)&0x8000)>>13);
	return q;
}

inline WORD CExecZ80::Sbc(WORD wArg1, WORD wArg2)
{
	int q   = wArg1 - wArg2 - (m_regF & C_FLAG);
	m_regF  = (((wArg1^q^wArg2)&0x1000)>>8)|((q>>16)&1)|((q&0x8000)>>8)|((q&65535)?0:Z_FLAG)|(((wArg1^wArg2)&(wArg2^q)&0x8000)>>13)|N_FLAG;
	return q;
}

inline WORD CExecZ80::Add(WORD wArg1, WORD wArg2)
{
	int q  = wArg1 + wArg2;
	m_regF = (m_regF&(S_FLAG|Z_FLAG|V_FLAG))|(((wArg1^q^wArg2)&0x1000)>>8)|((q>>16)&1);
	return q;
}

//
//  8-bit
//

inline void CExecZ80::Add(BYTE bArg)
{
#if 0
	int q = m_regA + bArg;
	m_regF=ZSTable[q&255]|((q&256)>>8)| ((m_regA^q^bArg)&H_FLAG)| (((bArg^m_regA^0x80)&(bArg^q)&0x80)>>5);
	m_regA=q;
#else
	int q = m_regA + bArg;
	m_regF = rgfAdd[m_regA][bArg];
	m_regA = q;
#endif
}

inline void CExecZ80::Adc(BYTE bArg)
{
#if 0
	int q  = m_regA + bArg + (m_regF & C_FLAG);
	m_regF = ZSTable[q & 255]|((q & 256)>>8)|((m_regA^q^bArg)&H_FLAG)|(((bArg^m_regA^0x80)&(bArg^q)&0x80)>>5);
	m_regA = q;
#else
	int q  = m_regA + bArg + (m_regF & C_FLAG);
	m_regF = rgfAddc[m_regA][bArg][m_regF & C_FLAG];
	m_regA = q;
#endif
}

inline void CExecZ80::Sub(BYTE bArg)
{
#if 0
	int q  = m_regA - bArg;
	m_regF = ZSTable[q&255]|((q&256)>>8)|N_FLAG|((m_regA^q^bArg)&H_FLAG)|(((bArg^m_regA)&(bArg^q)&0x80)>>5);
	m_regA = q;
#else
	int q  = m_regA - bArg;
	m_regF = rgfSub[m_regA][bArg];
	m_regA = q;
#endif
}

inline void CExecZ80::Sbc(BYTE bArg)
{
#if 0
	int q = m_regA - bArg - (m_regF & C_FLAG);
	m_regF=ZSTable[q&255]|((q&256)>>8)|N_FLAG|((m_regA^q^bArg)&H_FLAG)|(((bArg^m_regA)&(bArg^q)&0x80)>>5);
	m_regA=q;
#else
	int q = m_regA - bArg - (m_regF & C_FLAG);
	m_regF = rgfSubc[m_regA][bArg][m_regF & C_FLAG];
	m_regA = q;
#endif
}

inline void CExecZ80::And(BYTE bArg)
{
	m_regF = ZSPTable[m_regA &= bArg] | H_FLAG;
}

inline void CExecZ80::Or(BYTE bArg)
{
	m_regF = ZSPTable[m_regA |= bArg];
}

inline void CExecZ80::Xor(BYTE bArg)
{
	m_regF = ZSPTable[m_regA ^= bArg];
}

inline void CExecZ80::Cp(BYTE bArg)
{
#if 0
	int q = m_regA - bArg;
	m_regF = ZSTable[q&255]|((q&256)>>8)|N_FLAG|((m_regA^q^bArg)&H_FLAG)|(((bArg^m_regA)&(bArg^q)&0x80)>>5);
#else
	int q = m_regA - bArg;
	m_regF = rgfSub[m_regA][bArg];
#endif
}

inline BYTE CExecZ80::Inc(BYTE bArg)
{
#if 1
	++bArg;

	m_regF = (m_regF&C_FLAG)|ZSTable[bArg]|((bArg==0x80)?V_FLAG:0)|((bArg&0x0F)?0:H_FLAG);

	return bArg;
#else
	BYTE bArg2 = bArg++;
	m_regF = (m_regF & C_FLAG) | (rgfAdd[bArg][bArg2] & ~C_FLAG);
	return bArg2;
#endif
}

inline BYTE CExecZ80::Dec(BYTE bArg)
{
#if 1
	m_regF=(m_regF&C_FLAG)|N_FLAG|
		((bArg==0x80)?V_FLAG:0)|((bArg&0x0F)?0:H_FLAG);

	m_regF|=ZSTable[--bArg];

	return bArg;
#else
	BYTE bArg2 = bArg--;
	m_regF = (m_regF & C_FLAG) | (rgfSub[bArg][bArg2] & ~C_FLAG);
	return bArg2;
#endif
}

inline BYTE CExecZ80::Set(BYTE bArg, int nBit)
{
	return (bArg | (1 << nBit));
}

inline BYTE CExecZ80::Res(BYTE bArg, int nBit)
{
	return (bArg & ~(1 << nBit));
}

inline void CExecZ80::Bit(BYTE bArg, int nBit)
{
	m_regF = (m_regF & C_FLAG) | H_FLAG | ((bArg&(1<< nBit))? ((nBit==7)?S_FLAG:0):Z_FLAG);
}

inline BYTE CExecZ80::Rlc(BYTE bArg)
{
	int q  = bArg >> 7;
	bArg   = (bArg << 1) | q;
	m_regF = ZSPTable[bArg] | q;
	return bArg;
}

inline BYTE CExecZ80::Rrc(BYTE bArg)
{
	int q = bArg & 1;
	bArg = (bArg >> 1) | (q << 7);
	m_regF = ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Rl(BYTE bArg)
{
	int q=bArg>>7;
	bArg=(bArg<<1)|(m_regF&1);
	m_regF=ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Rr(BYTE bArg)
{
	int q = bArg&1;
	bArg=(bArg>>1)|(m_regF<<7);
	m_regF=ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Sll(BYTE bArg)
{
	int q = bArg>>7;
	bArg=(bArg<<1)|1;
	m_regF=ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Srl(BYTE bArg)
{
	int q = bArg&1;
	bArg>>=1;
	m_regF=ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Sla(BYTE bArg)
{
	int q = bArg>>7;
	bArg<<=1;
	m_regF=ZSPTable[bArg]|q;
	return bArg;
}

inline BYTE CExecZ80::Sra(BYTE bArg)
{
	int q = bArg&1;
	bArg = (bArg>>1)|(bArg&0x80);
	m_regF = ZSPTable[bArg]|q;
	return bArg;
}

#endif // _NEWZ80I_H_

